import dataclasses
from random import Random
from randovania.game_description.item import item_database
from typing import Dict, Iterator

import randovania
from randovania.game_description import default_database
from randovania.game_description.assignment import GateAssignment, PickupTarget
from randovania.game_description.default_database import default_prime2_memo_data
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches, ElevatorConnection
from randovania.game_description.resources.pickup_entry import PickupModel
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources, ResourceGain
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.node import TeleporterNode
from randovania.game_description.world.teleporter import Teleporter
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.games.prime import elevators
from randovania.games.prime.echoes_dol_patcher import EchoesDolPatchesData
from randovania.games.prime.patcher_file_lib import sky_temple_key_hint, item_names, pickup_exporter, hints, hint_lib, \
    credits_spoiler
from randovania.generator.item_pool import pickup_creator
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.base.base_configuration import BaseConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.lib.teleporters import TeleporterShuffleMode
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches
from randovania.layout.prime2.hint_configuration import HintConfiguration, SkyTempleKeyHintMode

_EASTER_EGG_RUN_VALIDATED_CHANCE = 1024
_EASTER_EGG_SHINY_MISSILE = 8192

_ENERGY_CONTROLLER_MAP_ASSET_IDS = [
    618058071,  # Agon EC
    724159530,  # Torvus EC
    988679813,  # Sanc EC
]
_ELEVATOR_ROOMS_MAP_ASSET_IDS = [
    # 0x529F0152,  # Sky Temple Energy Controller
    0xAE06A5D9,  # Sky Temple Gateway

    # cliff
    0x1C7CBD3E,  # agon
    0x92A2ADA3,  # Torvus
    0xFB9E9C00,  # Entrance
    0x74EFFB3C,  # Aerie
    0x932CB12E,  # Aerie Transport Station

    # sand
    0xEF5EA06C,  # Sanc
    0x8E9B3B3F,  # Torvus
    0x7E1BC16F,  # Entrance

    # swamp
    0x46B0EECF,  # Entrance
    0xE6B06473,  # Agon
    0x96DB1F15,  # Sanc

    # tg -> areas
    0x4B2A6FD3,  # Agon
    0x85E70805,  # Torvus
    0xE4229356,  # Sanc

    # tg -> gt
    0x79EFFD7D,
    0x65168477,
    0x84388E13,

    # gt -> tg
    0xA6D44A39,
    0x318EBBCD,
    0xB1B5308D,
]


def _create_spawn_point_field(patches: GamePatches,
                              resource_database: ResourceDatabase,
                              ) -> dict:
    capacities = [
        {
            "index": item.index,
            "amount": patches.starting_items.get(item, 0),
        }
        for item in resource_database.item
    ]

    return {
        "location": patches.starting_location.as_json,
        "amount": capacities,
        "capacity": capacities,
    }


def _pretty_name_for_elevator(game: RandovaniaGame,
                              world_list: WorldList,
                              original_teleporter_node: TeleporterNode,
                              connection: AreaLocation,
                              ) -> str:
    """
    Calculates the name the room that contains this elevator should have
    :param world_list:
    :param original_teleporter_node:
    :param connection:
    :return:
    """
    if original_teleporter_node.keep_name_when_vanilla:
        if original_teleporter_node.default_connection == connection:
            return world_list.nodes_to_area(original_teleporter_node).name

    return "Transport to {}".format(elevators.get_elevator_or_area_name(game, world_list, connection, False))


def _create_elevators_field(patches: GamePatches, game: GameDescription) -> list:
    """
    Creates the elevator entries in the patcher file
    :param patches:
    :param game:
    :return:
    """

    world_list = game.world_list
    nodes_by_teleporter = _get_nodes_by_teleporter_id(world_list)
    elevator_connection = patches.elevator_connection

    if len(elevator_connection) != len(nodes_by_teleporter):
        raise ValueError("Invalid elevator count. Expected {}, got {}.".format(
            len(nodes_by_teleporter), len(elevator_connection)
        ))

    elevator_fields = [
        {
            "instance_id": teleporter.instance_id,
            "origin_location": world_list.node_to_area_location(nodes_by_teleporter[teleporter]).as_json,
            "target_location": connection.as_json,
            "room_name": _pretty_name_for_elevator(game.game, world_list, nodes_by_teleporter[teleporter], connection)
        }
        for teleporter, connection in elevator_connection.items()
    ]

    return elevator_fields


def _get_nodes_by_teleporter_id(world_list: WorldList) -> Dict[Teleporter, TeleporterNode]:
    return {
        node.teleporter: node

        for node in world_list.all_nodes
        if isinstance(node, TeleporterNode) and node.editable
    }


def _create_translator_gates_field(gate_assignment: GateAssignment) -> list:
    """
    Creates the translator gate entries in the patcher file
    :param gate_assignment:
    :return:
    """
    return [
        {
            "gate_index": gate.index,
            "translator_index": translator.index,
        }
        for gate, translator in gate_assignment.items()
    ]


def _apply_translator_gate_patches(specific_patches: dict, elevator_shuffle_mode: TeleporterShuffleMode) -> None:
    """

    :param specific_patches:
    :param elevators:
    :return:
    """
    specific_patches["always_up_gfmc_compound"] = True
    specific_patches["always_up_torvus_temple"] = True
    specific_patches["always_up_great_temple"] = elevator_shuffle_mode != TeleporterShuffleMode.VANILLA


def _create_elevator_scan_port_patches(game:RandovaniaGame, world_list: WorldList, elevator_connection: ElevatorConnection,
                                       ) -> Iterator[dict]:
    nodes_by_teleporter_id = _get_nodes_by_teleporter_id(world_list)

    for teleporter, node in nodes_by_teleporter_id.items():
        if node.scan_asset_id is None:
            continue

        target_area_name = elevators.get_elevator_or_area_name(game, world_list, elevator_connection[teleporter], True)
        yield {
            "asset_id": node.scan_asset_id,
            "strings": [f"Access to &push;&main-color=#FF3333;{target_area_name}&pop; granted.", ""],
        }


def _logbook_title_string_patches():
    return [
        {
            "asset_id": 3271034066,
            "strings": [
                'Hints', 'Violet', 'Cobalt', 'Technology', 'Keys 1, 2, 3', 'Keys 7, 8, 9', 'Regular Hints',
                'Emerald', 'Amber', '&line-spacing=75;Flying Ing\nCache Hints', 'Keys 4, 5, 6', 'Keys 1, 2, 3',
                '&line-spacing=75;Torvus Energy\nController', 'Underground Tunnel', 'Training Chamber',
                'Catacombs', 'Gathering Hall', '&line-spacing=75;Fortress\nTransport\nAccess',
                '&line-spacing=75;Hall of Combat\nMastery', 'Main Gyro Chamber',
                '&line-spacing=75;Sanctuary\nEnergy\nController', 'Main Research', 'Watch Station',
                'Sanctuary Entrance', '&line-spacing=75;Transport to\nAgon Wastes', 'Mining Plaza',
                '&line-spacing=75;Agon Energy\nController', 'Portal Terminal', 'Mining Station B',
                'Mining Station A', 'Meeting Grounds', 'Path of Eyes', 'Path of Roots',
                '&line-spacing=75;Main Energy\nController', "Champions of Aether",
                '&line-spacing=75;Central\nMining\nStation', 'Main Reactor', 'Torvus Lagoon', 'Catacombs',
                'Sanctuary Entrance', "Dynamo Works", 'Storage Cavern A', 'Landing Site', 'Industrial Site',
                '&line-spacing=75;Sky Temple\nKey Hints', 'Keys 7, 8, 9', 'Keys 4, 5, 6', 'Sky Temple Key 1',
                'Sky Temple Key 2', 'Sky Temple Key 3', 'Sky Temple Key 4', 'Sky Temple Key 5',
                'Sky Temple Key 6', 'Sky Temple Key 7', 'Sky Temple Key 8', 'Sky Temple Key 9'
            ],
        }, {
            "asset_id": 2301408881,
            "strings": [
                'Research', 'Mechanisms', 'Luminoth Technology', 'Biology', 'GF Security', 'Vehicles',
                'Aether Studies', 'Aether', 'Dark Aether', 'Phazon', 'Sandgrass', 'Blueroot Tree',
                'Ing Webtrap',
                'Webling', 'U-Mos', 'Bladepod', 'Ing Storage', 'Flying Ing Cache', 'Torvus Bearerpod',
                'Agon Bearerpod', 'Ingworm Cache', 'Ingsphere Cache', 'Plantforms', 'Darklings',
                'GF Gate Mk VI',
                'GF Gate Mk VII', 'GF Lock Mk V', 'GF Defense Shield', 'Kinetic Orb Cannon', 'GF Bridge',
                "Samus's Gunship", 'GFS Tyr', 'Pirate Skiff', 'Visors', 'Weapon Systems', 'Armor',
                'Morph Ball Systems', 'Movement Systems', 'Beam Weapons', 'Scan Visor', 'Combat Visor',
                'Dark Visor',
                'Echo Visor', 'Morph Ball', 'Boost Ball', 'Spider Ball', 'Morph Ball Bomb', 'Power Bomb',
                'Dark Bomb', 'Light Bomb', 'Annihilator Bomb', 'Space Jump Boots', 'Screw Attack',
                'Gravity Boost',
                'Grapple Beam', 'Varia Suit', 'Dark Suit', 'Light Suit', 'Power Beam', 'Dark Beam',
                'Light Beam',
                'Annihilator Beam', 'Missile Launcher', 'Seeker Missile Launcher', 'Super Missile',
                'Sonic Boom',
                'Darkburst', 'Sunburst', 'Charge Beam', 'Missile Systems', 'Charge Combos', 'Morph Balls',
                'Bomb Systems', 'Miscellaneous', 'Dark Temple Keys', 'Bloatsac', 'Luminoth Technology',
                'Light Beacons', 'Light Crystals', 'Lift Crystals', 'Utility Crystals', 'Light Crystal',
                'Energized Crystal', 'Nullified Crystal', 'Super Crystal', 'Light Beacon', 'Energized Beacon',
                'Nullified Beacon', 'Super Beacon', 'Inactive Beacon', 'Dark Lift Crystal',
                'Light Lift Crystal',
                'Liftvine Crystal', 'Torvus Hanging Pod', 'Sentinel Crystal', 'Dark Sentinel Crystal',
                'Systems',
                'Bomb Slot', 'Spinner', 'Grapple Point', 'Spider Ball Track', 'Energy Tank',
                'Beam Ammo Expansion',
                'Missile Expansion', 'Dark Agon Keys', 'Dark Torvus Keys', 'Ing Hive Keys', 'Sky Temple Keys',
                'Temple Grounds', 'Sanctuary Fortress', 'Torvus Bog', 'Agon Wastes', 'Dark Agon Temple Key 1',
                'Dark Agon Temple Key 2', 'Dark Agon Temple Key 3', 'Dark Torvus Temple Key 1',
                'Dark Torvus Temple Key 2', 'Dark Torvus Temple Key 3', 'Ing Hive Temple Key 1',
                'Ing Hive Temple Key 2', 'Ing Hive Temple Key 3', 'Sky Temple Key 1', 'Sky Temple Key 2',
                'Sky Temple Key 3', 'Sky Temple Key 4', 'Sky Temple Key 5', 'Sky Temple Key 6',
                'Sky Temple Key 7',
                'Sky Temple Key 8', 'Sky Temple Key 9', 'Suit Expansions', 'Charge Combo', 'Ingclaw',
                'Dormant Ingclaw', 'Power Bomb Expansion', 'Energy Transfer Module', 'Cocoons',
                'Splinter Cocoon',
                'War Wasp Hive', 'Metroid Cocoon', 'Dark Aether', 'Aether', 'Dark Portal', 'Light Portal',
                'Energy Controller', 'Wall Jump Surface',
            ]
        },
    ]

def _akul_testament_string_patch():
    # update after each tournament! ordered from newest to oldest
    champs = [
        {
            "title": "2021 Champion",
            "name": "Dyceron"
        },
        {
            "title": "2020 Champion",
            "name": "Dyceron"
        }
    ]

    title = "Metroid Prime 2: Echoes Randomizer Tournament"
    champstring = '\n'.join([f'{champ["title"]}: {hint_lib.color_text(hint_lib.TextColor.PLAYER, champ["name"])}' for champ in champs])
    latest = champstring.partition("\n")[0]

    return [
        {
            "asset_id": 0x080BBD00,
            "strings": [
                'Luminoth Datapac translated.\n(Champions of Aether)',
                f"{title}\n\n{latest}",
                f"{title}\n\n{champstring}",
            ],
        },
    ]

def _create_string_patches(hint_config: HintConfiguration,
                           game: GameDescription,
                           all_patches: Dict[int, GamePatches],
                           area_namers: Dict[int, hint_lib.AreaNamer],
                           players_config: PlayersConfiguration,
                           rng: Random,
                           ) -> list:
    """

    :param hint_config:
    :param game:
    :param all_patches:
    :return:
    """
    patches = all_patches[players_config.player_index]
    string_patches = []

    string_patches.extend(_akul_testament_string_patch())

    # Location Hints
    string_patches.extend(
        hints.create_hints(all_patches, players_config, game.world_list, area_namers, rng)
    )

    # Sky Temple Keys
    stk_mode = hint_config.sky_temple_keys
    if stk_mode == SkyTempleKeyHintMode.DISABLED:
        string_patches.extend(sky_temple_key_hint.hide_hints())
    else:
        string_patches.extend(sky_temple_key_hint.create_hints(
            all_patches, players_config, game.resource_database,
            area_namers,
            stk_mode == SkyTempleKeyHintMode.HIDE_AREA))

    # Elevator Scans
    string_patches.extend(_create_elevator_scan_port_patches(game.game, game.world_list, patches.elevator_connection))

    string_patches.extend(_logbook_title_string_patches())

    

    return string_patches


def _create_starting_popup(layout_configuration: EchoesConfiguration,
                           resource_database: ResourceDatabase,
                           starting_items: CurrentResources) -> list:
    extra_items = item_names.additional_starting_items(layout_configuration, resource_database, starting_items)
    if extra_items:
        return [
            "Extra starting items:",
            ", ".join(extra_items)
        ]
    else:
        return []


def _simplified_memo_data() -> Dict[str, str]:
    result = pickup_exporter.GenericAcquiredMemo()
    result["Locked Power Bomb Expansion"] = ("Power Bomb Expansion acquired, "
                                             "but the main Power Bomb is required to use it.")
    result["Locked Missile Expansion"] = "Missile Expansion acquired, but the Missile Launcher is required to use it."
    result["Locked Seeker Launcher"] = "Seeker Launcher acquired, but the Missile Launcher is required to use it."
    return result


def _get_model_name_missing_backup():
    """
    A mapping of alternative model names if some models are missing.
    :return:
    """
    other_game = {
        PickupModel(RandovaniaGame.METROID_PRIME, "Charge Beam"): "ChargeBeam INCOMPLETE",
        PickupModel(RandovaniaGame.METROID_PRIME, "Super Missile"): "SuperMissile",
        PickupModel(RandovaniaGame.METROID_PRIME, "Scan Visor"): "ScanVisor INCOMPLETE",
        PickupModel(RandovaniaGame.METROID_PRIME, "Varia Suit"): "VariaSuit INCOMPLETE",
        PickupModel(RandovaniaGame.METROID_PRIME, "Gravity Suit"): "VariaSuit INCOMPLETE",
        PickupModel(RandovaniaGame.METROID_PRIME, "Phazon Suit"): "VariaSuit INCOMPLETE",
        # PickupModel(RandovaniaGame.PRIME1, "Morph Ball"): "MorphBall INCOMPLETE",
        PickupModel(RandovaniaGame.METROID_PRIME, "Morph Ball Bomb"): "MorphBallBomb",
        PickupModel(RandovaniaGame.METROID_PRIME, "Boost Ball"): "BoostBall",
        PickupModel(RandovaniaGame.METROID_PRIME, "Spider Ball"): "SpiderBall",
        PickupModel(RandovaniaGame.METROID_PRIME, "Power Bomb"): "PowerBomb",
        PickupModel(RandovaniaGame.METROID_PRIME, "Power Bomb Expansion"): "PowerBombExpansion",
        PickupModel(RandovaniaGame.METROID_PRIME, "Missile"): "MissileExpansionPrime1",
        PickupModel(RandovaniaGame.METROID_PRIME, "Grapple Beam"): "GrappleBeam",
        PickupModel(RandovaniaGame.METROID_PRIME, "Space Jump Boots"): "SpaceJumpBoots",
        PickupModel(RandovaniaGame.METROID_PRIME, "Energy Tank"): "EnergyTank",
    }
    return {
        f"{model.game.value}_{model.name}": name
        for model, name in other_game.items()
    }


def _get_model_mapping(randomizer_data: dict):
    jingles = {
        "SkyTempleKey": 2,
        "DarkTempleKey": 2,
        "MissileExpansion": 0,
        "PowerBombExpansion": 0,
        "DarkBeamAmmoExpansion": 0,
        "LightBeamAmmoExpansion": 0,
        "BeamAmmoExpansion": 0,
    }
    return EchoesModelNameMapping(
        index={
            entry["Name"]: entry["Index"]
            for entry in randomizer_data["ModelData"]
        },
        sound_index={
            "SkyTempleKey": 1,
            "DarkTempleKey": 1,
        },
        jingle_index={
            entry["Name"]: jingles.get(entry["Name"], 1)
            for entry in randomizer_data["ModelData"]
        },
    )


def create_patcher_file(description: LayoutDescription,
                        players_config: PlayersConfiguration,
                        cosmetic_patches: EchoesCosmeticPatches,
                        ) -> dict:
    """

    :param description:
    :param players_config:
    :param cosmetic_patches:
    :return:
    """
    preset = description.permalink.get_preset(players_config.player_index)
    configuration = preset.configuration
    patches = description.all_patches[players_config.player_index]
    rng = Random(description.permalink.seed_number)
    area_namers = {index: hint_lib.AreaNamer(default_database.game_description_for(preset.game).world_list)
                   for index, preset in description.permalink.presets.items()}

    game = default_database.game_description_for(RandovaniaGame.METROID_PRIME_ECHOES)

    result = {}
    _add_header_data_to_result(description, result)

    result["publisher_id"] = "0R"
    if configuration.menu_mod:
        result["publisher_id"] = "1R"

    result["convert_other_game_assets"] = cosmetic_patches.convert_other_game_assets
    result["credits"] = "\n\n\n\n\n" + credits_spoiler.prime_trilogy_credits(
        configuration.major_items_configuration,
        description.all_patches,
        players_config,
        area_namers,
        "&push;&main-color=#89D6FF;Major Item Locations&pop;",
        "&push;&main-color=#33ffd6;{}&pop;",
    )

    [item_category_visors] = [cat for cat in configuration.major_items_configuration.default_items.keys() if cat.name == "visor"]
    [item_category_beams] = [cat for cat in configuration.major_items_configuration.default_items.keys() if cat.name == "beam"]
    
    result["menu_mod"] = configuration.menu_mod
    result["dol_patches"] = EchoesDolPatchesData(
        energy_per_tank=configuration.energy_per_tank,
        beam_configuration=configuration.beam_configuration,
        safe_zone_heal_per_second=configuration.safe_zone.heal_per_second,
        user_preferences=cosmetic_patches.user_preferences,
        default_items={
            "visor": configuration.major_items_configuration.default_items[item_category_visors].name,
            "beam": configuration.major_items_configuration.default_items[item_category_beams].name,
        },
        unvisited_room_names=(configuration.elevators.can_use_unvisited_room_names
                              and cosmetic_patches.unvisited_room_names),
        teleporter_sounds=cosmetic_patches.teleporter_sounds or configuration.elevators.is_vanilla,
        dangerous_energy_tank=configuration.dangerous_energy_tank,
    ).as_json

    # Add Spawn Point
    result["spawn_point"] = _create_spawn_point_field(patches, game.resource_database)

    result["starting_popup"] = _create_starting_popup(configuration, game.resource_database, patches.starting_items)

    # Add the pickups
    result["pickups"] = _create_pickup_list(cosmetic_patches, configuration, game, patches, players_config, rng)

    # Add the elevators
    result["elevators"] = _create_elevators_field(patches, game)

    # Add translators
    result["translator_gates"] = _create_translator_gates_field(patches.translator_gates)

    # Scan hints
    result["string_patches"] = _create_string_patches(configuration.hints, game, description.all_patches,
                                                      area_namers, players_config, rng)

    # TODO: if we're starting at ship, needs to collect 9 sky temple keys and want item loss,
    # we should disable hive_chamber_b_post_state
    result["specific_patches"] = {
        "hive_chamber_b_post_state": True,
        "intro_in_post_state": True,
        "warp_to_start": configuration.warp_to_start,
        "credits_length": 75 if cosmetic_patches.speed_up_credits else 259,
        "disable_hud_popup": cosmetic_patches.disable_hud_popup,
        "pickup_map_icons": cosmetic_patches.pickup_markers,
        "full_map_at_start": cosmetic_patches.open_map,
        "dark_world_varia_suit_damage": configuration.varia_suit_damage,
        "dark_world_dark_suit_damage": configuration.dark_suit_damage,
    }

    result["logbook_patches"] = [
        {"asset_id": 25, "connections": [81, 166, 195], },
        {"asset_id": 38, "connections": [4, 33, 120, 251, 364], },
        {"asset_id": 60, "connections": [38, 74, 154, 196], },
        {"asset_id": 74, "connections": [59, 75, 82, 102, 260], },
        {"asset_id": 81, "connections": [148, 151, 156], },
        {"asset_id": 119, "connections": [60, 254, 326], },
        {"asset_id": 124, "connections": [35, 152, 355], },
        {"asset_id": 129, "connections": [29, 118, 367], },
        {"asset_id": 154, "connections": [169, 200, 228, 243, 312, 342], },
        {"asset_id": 166, "connections": [45, 303, 317], },
        {"asset_id": 194, "connections": [1, 6], },
        {"asset_id": 195, "connections": [159, 221, 231], },
        {"asset_id": 196, "connections": [17, 19, 23, 162, 183, 379], },
        {"asset_id": 233, "connections": [58, 191, 373], },
        {"asset_id": 241, "connections": [223, 284], },
        {"asset_id": 254, "connections": [129, 233, 319], },
        {"asset_id": 318, "connections": [119, 216, 277, 343], },
        {"asset_id": 319, "connections": [52, 289, 329], },
        {"asset_id": 326, "connections": [124, 194, 241, 327], },
        {"asset_id": 327, "connections": [46, 275], },
    ]

    if not configuration.elevators.is_vanilla and (cosmetic_patches.unvisited_room_names
                                                   and configuration.elevators.can_use_unvisited_room_names):
        exclude_map_ids = _ELEVATOR_ROOMS_MAP_ASSET_IDS
    else:
        exclude_map_ids = []
    result["maps_to_always_reveal"] = _ENERGY_CONTROLLER_MAP_ASSET_IDS
    result["maps_to_never_reveal"] = exclude_map_ids

    _apply_translator_gate_patches(result["specific_patches"], configuration.elevators.mode)

    return result


def _create_pickup_list(cosmetic_patches: EchoesCosmeticPatches, configuration: BaseConfiguration,
                        game: GameDescription,
                        patches: GamePatches, players_config: PlayersConfiguration,
                        rng: Random):
    useless_target = PickupTarget(pickup_creator.create_echoes_useless_pickup(game.resource_database),
                                  players_config.player_index)

    if cosmetic_patches.disable_hud_popup:
        memo_data = _simplified_memo_data()
    else:
        memo_data = default_prime2_memo_data()

    pickup_list = pickup_exporter.export_all_indices(
        patches,
        useless_target,
        game.world_list,
        rng,
        configuration.pickup_model_style,
        configuration.pickup_model_data_source,
        exporter=pickup_exporter.create_pickup_exporter(game, memo_data, players_config),
        visual_etm=pickup_creator.create_visual_etm(),
    )

    return [
        echoes_pickup_details_to_patcher(details, rng)
        for details in pickup_list
    ]


def _add_header_data_to_result(description: LayoutDescription, result: dict) -> None:
    result["permalink"] = "-permalink-"
    result["seed_hash"] = f"- {description.shareable_word_hash} ({description.shareable_hash})"
    result["shareable_hash"] = description.shareable_hash
    result["shareable_word_hash"] = description.shareable_word_hash
    result["randovania_version"] = randovania.VERSION


@dataclasses.dataclass(frozen=True)
class EchoesModelNameMapping:
    index: Dict[str, int]
    sound_index: Dict[str, int]  # 1 for keys, 0 otherwise
    jingle_index: Dict[str, int]  # 2 for keys, 1 for major items, 0 otherwise


def _create_pickup_resources_for(resources: ResourceGain):
    return [
        {
            "index": resource.index,
            "amount": quantity
        }
        for resource, quantity in resources
        if quantity > 0 and resource.resource_type == ResourceType.ITEM
    ]


def echoes_pickup_details_to_patcher(details: pickup_exporter.ExportedPickupDetails, rng: Random) -> dict:
    if details.model.game == RandovaniaGame.METROID_PRIME_ECHOES:
        model_name = details.model.name
    else:
        model_name = f"{details.model.game.value}_{details.model.name}"

    if model_name == "MissileExpansion" and rng.randint(0, _EASTER_EGG_SHINY_MISSILE) == 0:
        # If placing a missile expansion model, replace with Dark Missile Trooper model with a 1/8192 chance
        model_name = "MissileExpansionPrime1"

    hud_text = details.hud_text
    if hud_text == ["Energy Transfer Module acquired!"] and (
            rng.randint(0, _EASTER_EGG_RUN_VALIDATED_CHANCE) == 0):
        hud_text = ["Run validated!"]

    return {
        "pickup_index": details.index.index,
        "resources": _create_pickup_resources_for(details.conditional_resources[0].resources),
        "conditional_resources": [
            {
                "item": conditional.item.index,
                "resources": _create_pickup_resources_for(conditional.resources),
            }
            for conditional in details.conditional_resources[1:]
        ],
        "convert": [
            {
                "from_item": conversion.source.index,
                "to_item": conversion.target.index,
                "clear_source": conversion.clear_source,
                "overwrite_target": conversion.overwrite_target,
            }
            for conversion in details.conversion
        ],
        "hud_text": hud_text,
        "scan": details.scan_text,
        "model_name": model_name,
    }


def adjust_model_name(patcher_data: dict, randomizer_data: dict):
    mapping = _get_model_mapping(randomizer_data)
    backup = _get_model_name_missing_backup()

    for pickup in patcher_data["pickups"]:
        model_name = pickup.pop("model_name")
        if model_name not in mapping.index:
            model_name = backup.get(model_name, "EnergyTransferModule")

        pickup["model_index"] = mapping.index[model_name]
        pickup["sound_index"] = mapping.sound_index.get(model_name, 0)
        pickup["jingle_index"] = mapping.jingle_index.get(model_name, 0)



from PySide2 import QtWidgets

from randovania.gui.generated.scroll_label_dialog_ui import Ui_ScrollLabelDialog
from randovania.gui.lib import common_qt_lib


class ScrollLabelDialog(QtWidgets.QDialog, Ui_ScrollLabelDialog):
    def __init__(self, text: str, title: str, parent: QtWidgets.QWidget):
        super().__init__(parent)
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)
        self.setWindowTitle(title)
        self.label.setText(text)

        self.button_box.accepted.connect(self.accept)



import dataclasses

from PySide2.QtWidgets import QWidget

from randovania.gui.dialog.base_cosmetic_patches_dialog import BaseCosmeticPatchesDialog
from randovania.gui.generated.corruption_cosmetic_patches_dialog_ui import Ui_CorruptionCosmeticPatchesDialog
from randovania.gui.generated.prime_cosmetic_patches_dialog_ui import Ui_PrimeCosmeticPatchesDialog
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches
from randovania.layout.prime3.corruption_cosmetic_patches import CorruptionCosmeticPatches, CorruptionSuit


class CorruptionCosmeticPatchesDialog(BaseCosmeticPatchesDialog, Ui_CorruptionCosmeticPatchesDialog):
    _cosmetic_patches: CorruptionCosmeticPatches

    def __init__(self, parent: QWidget, current: CorruptionCosmeticPatches):
        super().__init__(parent)
        self.setupUi(self)
        self._cosmetic_patches = current

        for player_suit in CorruptionSuit:
            self.suit_combo.addItem(player_suit.name, player_suit)

        self.on_new_cosmetic_patches(current)
        self.connect_signals()

    def connect_signals(self):
        super().connect_signals()

        self.random_door_colors_check.stateChanged.connect(self._persist_option_then_notify("random_door_colors"))
        self.random_welding_colors_check.stateChanged.connect(self._persist_option_then_notify("random_welding_colors"))
        self.suit_combo.currentIndexChanged.connect(self._on_suit_update)

    def on_new_cosmetic_patches(self, patches: CorruptionCosmeticPatches):
        self.random_door_colors_check.setChecked(patches.random_door_colors)
        self.random_welding_colors_check.setChecked(patches.random_welding_colors)
        self.suit_combo.setCurrentIndex(self.suit_combo.findData(patches.player_suit))

    def _persist_option_then_notify(self, attribute_name: str):
        def persist(value: int):
            self._cosmetic_patches = dataclasses.replace(
                self._cosmetic_patches,
                **{attribute_name: bool(value)}
            )

        return persist

    def _on_suit_update(self):
        self._cosmetic_patches = dataclasses.replace(
            self._cosmetic_patches,
            player_suit=self.suit_combo.currentData(),
        )

    @property
    def cosmetic_patches(self) -> CorruptionCosmeticPatches:
        return self._cosmetic_patches

    def reset(self):
        self.on_new_cosmetic_patches(CorruptionCosmeticPatches())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'permalink_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PermalinkDialog(object):
    def setupUi(self, PermalinkDialog):
        if not PermalinkDialog.objectName():
            PermalinkDialog.setObjectName(u"PermalinkDialog")
        PermalinkDialog.resize(539, 117)
        self.gridLayout = QGridLayout(PermalinkDialog)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.accept_button = QPushButton(PermalinkDialog)
        self.accept_button.setObjectName(u"accept_button")

        self.gridLayout.addWidget(self.accept_button, 6, 0, 1, 1)

        self.permalink_edit = QLineEdit(PermalinkDialog)
        self.permalink_edit.setObjectName(u"permalink_edit")

        self.gridLayout.addWidget(self.permalink_edit, 4, 0, 1, 3)

        self.cancel_button = QPushButton(PermalinkDialog)
        self.cancel_button.setObjectName(u"cancel_button")

        self.gridLayout.addWidget(self.cancel_button, 6, 2, 1, 1)

        self.paste_button = QPushButton(PermalinkDialog)
        self.paste_button.setObjectName(u"paste_button")

        self.gridLayout.addWidget(self.paste_button, 1, 2, 1, 1)

        self.description_label = QLabel(PermalinkDialog)
        self.description_label.setObjectName(u"description_label")

        self.gridLayout.addWidget(self.description_label, 1, 0, 1, 2)

        self.import_error_label = QLabel(PermalinkDialog)
        self.import_error_label.setObjectName(u"import_error_label")
        self.import_error_label.setWordWrap(True)

        self.gridLayout.addWidget(self.import_error_label, 5, 0, 1, 3)


        self.retranslateUi(PermalinkDialog)

        QMetaObject.connectSlotsByName(PermalinkDialog)
    # setupUi

    def retranslateUi(self, PermalinkDialog):
        PermalinkDialog.setWindowTitle(QCoreApplication.translate("PermalinkDialog", u"Import permalink", None))
        self.accept_button.setText(QCoreApplication.translate("PermalinkDialog", u"Accept", None))
        self.permalink_edit.setPlaceholderText(QCoreApplication.translate("PermalinkDialog", u"Permalink", None))
        self.cancel_button.setText(QCoreApplication.translate("PermalinkDialog", u"Cancel", None))
        self.paste_button.setText(QCoreApplication.translate("PermalinkDialog", u"Paste", None))
        self.description_label.setText(QCoreApplication.translate("PermalinkDialog", u"<html><head/><body><p>Import a permalink that was shared by someone else.</p></body></html>", None))
        self.import_error_label.setText("")
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'auto_tracker_window.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_AutoTrackerWindow(object):
    def setupUi(self, AutoTrackerWindow):
        if not AutoTrackerWindow.objectName():
            AutoTrackerWindow.setObjectName(u"AutoTrackerWindow")
        AutoTrackerWindow.resize(162, 199)
        self.action_theme_2d_style = QAction(AutoTrackerWindow)
        self.action_theme_2d_style.setObjectName(u"action_theme_2d_style")
        self.action_theme_2d_style.setCheckable(True)
        self.action_theme_game_icons = QAction(AutoTrackerWindow)
        self.action_theme_game_icons.setObjectName(u"action_theme_game_icons")
        self.action_theme_game_icons.setCheckable(True)
        self.actionasdf = QAction(AutoTrackerWindow)
        self.actionasdf.setObjectName(u"actionasdf")
        self.action_force_update = QAction(AutoTrackerWindow)
        self.action_force_update.setObjectName(u"action_force_update")
        self.centralwidget = QWidget(AutoTrackerWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout = QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.connection_status_label = QLabel(self.centralwidget)
        self.connection_status_label.setObjectName(u"connection_status_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.connection_status_label.sizePolicy().hasHeightForWidth())
        self.connection_status_label.setSizePolicy(sizePolicy)

        self.verticalLayout.addWidget(self.connection_status_label)


        self.gridLayout.addLayout(self.verticalLayout, 1, 0, 1, 1)

        self.inventory_group = QGroupBox(self.centralwidget)
        self.inventory_group.setObjectName(u"inventory_group")
        self.inventory_group.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.inventory_layout = QGridLayout(self.inventory_group)
        self.inventory_layout.setObjectName(u"inventory_layout")

        self.gridLayout.addWidget(self.inventory_group, 0, 0, 1, 1)

        AutoTrackerWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(AutoTrackerWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 162, 22))
        self.menu_options = QMenu(self.menubar)
        self.menu_options.setObjectName(u"menu_options")
        self.menu_tracker = QMenu(self.menu_options)
        self.menu_tracker.setObjectName(u"menu_tracker")
        self.menu_backend = QMenu(self.menu_options)
        self.menu_backend.setObjectName(u"menu_backend")
        AutoTrackerWindow.setMenuBar(self.menubar)

        self.menubar.addAction(self.menu_options.menuAction())
        self.menu_options.addAction(self.menu_tracker.menuAction())
        self.menu_options.addAction(self.menu_backend.menuAction())
        self.menu_options.addAction(self.action_force_update)

        self.retranslateUi(AutoTrackerWindow)

        QMetaObject.connectSlotsByName(AutoTrackerWindow)
    # setupUi

    def retranslateUi(self, AutoTrackerWindow):
        AutoTrackerWindow.setWindowTitle(QCoreApplication.translate("AutoTrackerWindow", u"Auto Tracker", None))
        self.action_theme_2d_style.setText(QCoreApplication.translate("AutoTrackerWindow", u"2D Style", None))
        self.action_theme_game_icons.setText(QCoreApplication.translate("AutoTrackerWindow", u"Game Icons (Echoes only)", None))
        self.actionasdf.setText(QCoreApplication.translate("AutoTrackerWindow", u"asdf", None))
        self.action_force_update.setText(QCoreApplication.translate("AutoTrackerWindow", u"Force Update", None))
        self.connection_status_label.setText(QCoreApplication.translate("AutoTrackerWindow", u"Connection Status", None))
        self.inventory_group.setTitle(QCoreApplication.translate("AutoTrackerWindow", u"Inventory", None))
        self.menu_options.setTitle(QCoreApplication.translate("AutoTrackerWindow", u"Options", None))
        self.menu_tracker.setTitle(QCoreApplication.translate("AutoTrackerWindow", u"Tracker", None))
        self.menu_backend.setTitle(QCoreApplication.translate("AutoTrackerWindow", u"Configure backend", None))
    # retranslateUi




import typing

from PySide2 import QtWidgets

from randovania.gui.generated.preset_prime_patches_ui import Ui_PresetPrimePatches
from randovania.gui.lib import signal_handling
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset
from randovania.layout.prime1.prime_configuration import LayoutCutsceneMode

_FIELDS = [
    "warp_to_start",
    "main_plaza_door",
    "backwards_frigate",
    "backwards_labs",
    "backwards_upper_mines",
    "backwards_lower_mines",
    "phazon_elite_without_dynamo",
    "qol_game_breaking",
    "qol_pickup_scans",
    "small_samus",
]


class PresetPrimePatches(PresetTab, Ui_PresetPrimePatches):

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        self.setupUi(self)

        self.description_label.setText(self.description_label.text().replace("color:#0000ff;", ""))

        # Signals
        self.cutscene_combo.setItemData(0, LayoutCutsceneMode.ORIGINAL)
        self.cutscene_combo.setItemData(1, LayoutCutsceneMode.COMPETITIVE)
        self.cutscene_combo.setItemData(2, LayoutCutsceneMode.MINOR)
        self.cutscene_combo.setItemData(3, LayoutCutsceneMode.MAJOR)
        signal_handling.on_combo(self.cutscene_combo, self._on_cutscene_changed)
        for f in _FIELDS:
            self._add_persist_option(getattr(self, f"{f}_check"), f)

    @property
    def uses_patches_tab(self) -> bool:
        return True

    def _add_persist_option(self, check: QtWidgets.QCheckBox, attribute_name: str):
        def persist(value: bool):
            with self._editor as editor:
                editor.set_configuration_field(attribute_name, value)

        signal_handling.on_checked(check, persist)

    def _on_cutscene_changed(self, value: LayoutCutsceneMode):
        with self._editor as editor:
            editor.set_configuration_field("qol_cutscenes", value)

    def on_preset_changed(self, preset: Preset):
        config = preset.configuration
        for f in _FIELDS:
            typing.cast(QtWidgets.QCheckBox, getattr(self, f"{f}_check")).setChecked(getattr(config, f))
        signal_handling.combo_set_to_value(self.cutscene_combo, config.qol_cutscenes)



from PySide2 import QtWidgets

from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.gui.dialog.base_cosmetic_patches_dialog import BaseCosmeticPatchesDialog
from randovania.gui.dialog.corruption_cosmetic_patches_dialog import CorruptionCosmeticPatchesDialog
from randovania.gui.dialog.echoes_cosmetic_patches_dialog import EchoesCosmeticPatchesDialog
from randovania.gui.dialog.prime_cosmetic_patches_dialog import PrimeCosmeticPatchesDialog
from randovania.gui.lib.window_manager import WindowManager
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.game_to_class import AnyCosmeticPatches

COSMETIC_DIALOG_FOR_GAME = {
    RandovaniaGame.METROID_PRIME: PrimeCosmeticPatchesDialog,
    RandovaniaGame.METROID_PRIME_ECHOES: EchoesCosmeticPatchesDialog,
    RandovaniaGame.METROID_PRIME_CORRUPTION: CorruptionCosmeticPatchesDialog,
}


def create_dialog_for_cosmetic_patches(
        parent: QtWidgets.QWidget,
        initial_patches: AnyCosmeticPatches,
) -> BaseCosmeticPatchesDialog:
    game = initial_patches.game()
    dialog_class = COSMETIC_DIALOG_FOR_GAME[game]
    return dialog_class(parent, initial_patches)


def prime1_preset_tabs(editor: PresetEditor, window_manager: WindowManager):
    game_enum = editor.game
    game_description = default_database.game_description_for(game_enum)

    from randovania.gui.preset_settings.trick_level_tab import PresetTrickLevel
    from randovania.gui.preset_settings.patcher_energy_tab import PresetPatcherEnergy
    from randovania.gui.preset_settings.elevators_tab import PresetElevators
    from randovania.gui.preset_settings.starting_area_tab import PresetStartingArea
    from randovania.gui.preset_settings.logic_damage_tab import PresetLogicDamage
    from randovania.gui.preset_settings.prime1.prime_goal_tab import PresetPrimeGoal
    from randovania.gui.preset_settings.prime1.prime_patches_tab import PresetPrimePatches
    from randovania.gui.preset_settings.location_pool_tab import PresetLocationPool
    from randovania.gui.preset_settings.metroid_item_pool_tab import MetroidPresetItemPool
    return [
        PresetTrickLevel(editor, game_description, window_manager),
        PresetPatcherEnergy(editor, game_enum),
        PresetElevators(editor, game_description),
        PresetStartingArea(editor, game_description),
        PresetLogicDamage(editor),
        PresetPrimeGoal(editor),
        PresetPrimePatches(editor),
        PresetLocationPool(editor, game_description),
        MetroidPresetItemPool(editor),
    ]


def prime2_preset_tabs(editor: PresetEditor, window_manager: WindowManager):
    game_enum = editor.game
    game_description = default_database.game_description_for(game_enum)

    from randovania.gui.preset_settings.trick_level_tab import PresetTrickLevel
    from randovania.gui.preset_settings.patcher_energy_tab import PresetPatcherEnergy
    from randovania.gui.preset_settings.elevators_tab import PresetElevators
    from randovania.gui.preset_settings.starting_area_tab import PresetStartingArea
    from randovania.gui.preset_settings.logic_damage_tab import PresetLogicDamage
    from randovania.gui.preset_settings.prime2.echoes_goal_tab import PresetEchoesGoal
    from randovania.gui.preset_settings.prime2.echoes_hints_tab import PresetEchoesHints
    from randovania.gui.preset_settings.prime2.echoes_translators_tab import PresetEchoesTranslators
    from randovania.gui.preset_settings.prime2.echoes_beam_configuration_tab import PresetEchoesBeamConfiguration
    from randovania.gui.preset_settings.prime2.echoes_patches_tab import PresetEchoesPatches
    from randovania.gui.preset_settings.location_pool_tab import PresetLocationPool
    from randovania.gui.preset_settings.prime2.echoes_item_pool_tab import EchoesPresetItemPool
    return [
        PresetTrickLevel(editor, game_description, window_manager),
        PresetPatcherEnergy(editor, game_enum),
        PresetElevators(editor, game_description),
        PresetStartingArea(editor, game_description),
        PresetLogicDamage(editor),
        PresetEchoesGoal(editor),
        PresetEchoesHints(editor),
        PresetEchoesTranslators(editor),
        PresetEchoesBeamConfiguration(editor),
        PresetEchoesPatches(editor),
        PresetLocationPool(editor, game_description),
        EchoesPresetItemPool(editor),
    ]


def prime3_preset_tabs(editor: PresetEditor, window_manager: WindowManager):
    game_enum = editor.game
    game_description = default_database.game_description_for(game_enum)

    from randovania.gui.preset_settings.trick_level_tab import PresetTrickLevel
    from randovania.gui.preset_settings.patcher_energy_tab import PresetPatcherEnergy
    from randovania.gui.preset_settings.elevators_tab import PresetElevators
    from randovania.gui.preset_settings.starting_area_tab import PresetStartingArea
    from randovania.gui.preset_settings.logic_damage_tab import PresetLogicDamage
    from randovania.gui.preset_settings.location_pool_tab import PresetLocationPool
    from randovania.gui.preset_settings.metroid_item_pool_tab import MetroidPresetItemPool
    return [
        PresetTrickLevel(editor, game_description, window_manager),
        PresetPatcherEnergy(editor, game_enum),
        PresetElevators(editor, game_description),
        PresetStartingArea(editor, game_description),
        PresetLogicDamage(editor),
        PresetLocationPool(editor, game_description),
        MetroidPresetItemPool(editor),
    ]


def super_metroid_preset_tabs(editor: PresetEditor, window_manager: WindowManager):
    game_enum = editor.game
    game_description = default_database.game_description_for(game_enum)
    from randovania.gui.preset_settings.trick_level_tab import PresetTrickLevel
    from randovania.gui.preset_settings.starting_area_tab import PresetStartingArea
    from randovania.gui.preset_settings.location_pool_tab import PresetLocationPool
    # from randovania.gui.preset_settings.metroid_item_pool_tab import MetroidPresetItemPool

    return [
        PresetTrickLevel(editor, game_description, window_manager),
        PresetStartingArea(editor, game_description),
        PresetLocationPool(editor, game_description),
        # MetroidPresetItemPool(editor),
    ]


TAB_PROVIDER_FOR_GAME = {
    RandovaniaGame.METROID_PRIME: prime1_preset_tabs,
    RandovaniaGame.METROID_PRIME_ECHOES: prime2_preset_tabs,
    RandovaniaGame.METROID_PRIME_CORRUPTION: prime3_preset_tabs,
    RandovaniaGame.SUPER_METROID: super_metroid_preset_tabs,
}


def preset_editor_tabs_for(editor: PresetEditor, window_manager: WindowManager):
    return TAB_PROVIDER_FOR_GAME[editor.game](editor, window_manager)



import traceback
from typing import Optional


class ExportFailure(Exception):
    def __init__(self, reason: str, output: Optional[str]):
        super().__init__(reason)
        self.output = output

    def detailed_text(self) -> str:
        result = []
        if self.output is not None:
            result.append(self.output)
        if self.__traceback__ is not None:
            result.extend(traceback.format_tb(self.__traceback__))
        return "\n".join(result)



import dataclasses
import json
from pathlib import Path
import uuid
from randovania.game_description.item.item_category import ItemCategory
from unittest.mock import MagicMock

import pytest

import randovania.games.patchers.claris_patcher
from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.item.item_database import ItemDatabase
from randovania.game_description.resources.pickup_entry import PickupEntry, PickupModel
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games import default_data
from randovania.games.game import RandovaniaGame
from randovania.interface_common.preset_manager import PresetManager
from randovania.layout.preset import Preset
from randovania.layout.prime2.echoes_configuration import EchoesConfiguration


@pytest.fixture(scope="session")
def test_files_dir() -> Path:
    return Path(__file__).parent.joinpath("test_files")


@pytest.fixture(scope="session")
def echo_tool(request, test_files_dir) -> Path:
    if request.config.option.skip_echo_tool:
        pytest.skip()
    return test_files_dir.joinpath("echo_tool.py")


@pytest.fixture()
def preset_manager(tmp_path) -> PresetManager:
    return PresetManager(tmp_path.joinpath("presets"))


@pytest.fixture(scope="session")
def default_preset() -> Preset:
    return PresetManager(None).default_preset.get_preset()

@pytest.fixture(scope="session")
def customized_preset(default_preset) -> Preset:
    return Preset(
        name="{} Custom".format(default_preset.name),
        description="A customized preset.",
        uuid=uuid.uuid4(),
        base_preset_uuid=default_preset.uuid,
        game=default_preset.game,
        configuration=default_preset.configuration
    )

@pytest.fixture(scope="session")
def default_echoes_preset() -> Preset:
    return PresetManager(None).default_preset_for_game(RandovaniaGame.METROID_PRIME_ECHOES).get_preset()


@pytest.fixture(scope="session")
def default_layout_configuration() -> EchoesConfiguration:
    return PresetManager(None).default_preset.get_preset().configuration


@pytest.fixture(scope="session")
def prime1_resource_database() -> ResourceDatabase:
    return default_database.resource_database_for(RandovaniaGame.METROID_PRIME)


@pytest.fixture(scope="session")
def echoes_resource_database() -> ResourceDatabase:
    return default_database.resource_database_for(RandovaniaGame.METROID_PRIME_ECHOES)


@pytest.fixture(scope="session")
def echoes_item_database() -> ItemDatabase:
    return default_database.item_database_for_game(RandovaniaGame.METROID_PRIME_ECHOES)


@pytest.fixture(scope="session")
def echoes_game_data() -> dict:
    return default_data.read_json_then_binary(RandovaniaGame.METROID_PRIME_ECHOES)[1]


@pytest.fixture(scope="session")
def echoes_game_description(echoes_game_data) -> GameDescription:
    return default_database.game_description_for(RandovaniaGame.METROID_PRIME_ECHOES)


@pytest.fixture(scope="session")
def corruption_game_data() -> dict:
    return default_data.read_json_then_binary(RandovaniaGame.METROID_PRIME_CORRUPTION)[1]


@pytest.fixture(scope="session")
def corruption_game_description(corruption_game_data) -> GameDescription:
    return default_database.game_description_for(RandovaniaGame.METROID_PRIME_CORRUPTION)


@pytest.fixture(scope="session")
def randomizer_data() -> dict:
    return randovania.games.patchers.claris_patcher.decode_randomizer_data()


@pytest.fixture(params=RandovaniaGame)
def game_enum(request) -> RandovaniaGame:
    return request.param


@pytest.fixture()
def generic_item_category() -> ItemCategory:
    return ItemCategory(
        name="generic",
        long_name="Generic Item Category",
        hint_details=("an ", "unspecified item"),
        is_major=False
    )

@pytest.fixture()
def blank_pickup(echoes_item_database) -> PickupEntry:
    return PickupEntry(
        name="Blank Pickup",
        model=PickupModel(
            game=RandovaniaGame.METROID_PRIME_ECHOES,
            name="EnergyTransferModule",
        ),
        item_category=echoes_item_database.item_categories["suit"],
        broad_category=echoes_item_database.item_categories["life_support"],
        progression=(),
        resource_lock=None,
        unlocks_resource=False,
    )


@pytest.fixture()
def small_echoes_game_description(test_files_dir) -> GameDescription:
    from randovania.game_description import data_reader
    with test_files_dir.joinpath("prime2_small.json").open("r") as small_game:
        return data_reader.decode_data(json.load(small_game))


class DataclassTestLib:
    def mock_dataclass(self, obj) -> MagicMock:
        return MagicMock(spec=[field.name for field in dataclasses.fields(obj)])


@pytest.fixture()
def dataclass_test_lib() -> DataclassTestLib:
    return DataclassTestLib()


@pytest.fixture()
def empty_patches() -> GamePatches:
    return GamePatches(0, {}, {}, {}, {}, {}, {}, None, {})


def pytest_addoption(parser):
    parser.addoption('--skip-generation-tests', action='store_true', dest="skip_generation_tests",
                     default=False, help="Skips running layout generation tests")
    parser.addoption('--skip-resolver-tests', action='store_true', dest="skip_resolver_tests",
                     default=False, help="Skips running validation tests")
    parser.addoption('--skip-gui-tests', action='store_true', dest="skip_gui_tests",
                     default=False, help="Skips running GUI tests")
    parser.addoption('--skip-echo-tool', action='store_true', dest="skip_echo_tool",
                     default=False, help="Skips running tests that uses the echo tool")


try:
    import pytestqt


    @pytest.fixture()
    def skip_qtbot(request, qtbot):
        if request.config.option.skip_gui_tests:
            pytest.skip()
        return qtbot

except ImportError:
    @pytest.fixture()
    def skip_qtbot(request):
        pytest.skip()


def pytest_configure(config):
    if config.option.skip_generation_tests:
        setattr(config.option, 'markexpr', 'not skip_generation_tests')

    if config.option.skip_resolver_tests:
        setattr(config.option, 'markexpr', 'not skip_resolver_tests')

    if config.option.skip_gui_tests:
        setattr(config.option, 'markexpr', 'not skip_gui_tests')



from setuptools import setup
from setuptools.command.build_py import build_py
from wheel.bdist_wheel import bdist_wheel

try:
    from pyqt_distutils.build_ui import build_ui
except ModuleNotFoundError:
    build_ui = None


class BuildPyCommand(build_py):
    def run(self):
        self.run_command('build_ui')
        super().run()


class BDistWheelCommand(bdist_wheel):
    def run(self):
        self.run_command('build_ui')
        super().run()


def version_scheme(version):
    import setuptools_scm.version
    if version.exact:
        return setuptools_scm.version.guess_next_simple_semver(
            version.tag, retain=setuptools_scm.version.SEMVER_LEN, increment=False
        )
    else:
        if version.branch != "stable":
            return version.format_next_version(
                setuptools_scm.version.guess_next_simple_semver, retain=setuptools_scm.version.SEMVER_MINOR
            )
        else:
            return version.format_next_version(
                setuptools_scm.version.guess_next_simple_semver, retain=setuptools_scm.version.SEMVER_PATCH
            )


setup(
    use_scm_version={
        "version_scheme": version_scheme,
        "local_scheme": "no-local-version",
        "write_to": "randovania/version.py",
    },
    cmdclass={
        "build_ui": build_ui,
        "build_py": BuildPyCommand,
        'bdist_wheel': BDistWheelCommand,
    },
)



import collections
import dataclasses
import functools
from randovania.gui.preset_settings.location_pool_row_widget import LocationPoolRowWidget
import re
from typing import Dict

from PySide2 import QtWidgets
from PySide2.QtWidgets import QFrame, QGraphicsOpacityEffect, QSizePolicy, QSpacerItem

from randovania.game_description.game_description import GameDescription
from randovania.game_description.world.node import Node, PickupNode
from randovania.games.game import RandovaniaGame
from randovania.gui.generated.preset_location_pool_ui import Ui_PresetLocationPool
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.foldable import Foldable
from randovania.gui.lib.area_list_helper import AreaListHelper, dark_world_flags
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.available_locations import RandomizationMode
from randovania.layout.preset import Preset


class PresetLocationPool(PresetTab, Ui_PresetLocationPool, AreaListHelper):

    _starting_location_for_world: Dict[str, QtWidgets.QCheckBox]
    _starting_location_for_area: Dict[int, QtWidgets.QCheckBox]
    _row_widget_for_node: Dict[Node, LocationPoolRowWidget]
    _during_batch_update: bool

    def __init__(self, editor: PresetEditor, game: GameDescription):
        super().__init__(editor)
        self.setupUi(self)
        self.game_description = game
        self._during_batch_update = False

        self._row_widget_for_node = {}
        
        self.check_major_minor.stateChanged.connect(self._on_update_randomization_mode)

        world_list = self.game_description.world_list
        
        nodes_by_world = collections.defaultdict(list)
        node_names = {}
        pickup_match = re.compile(r"Pickup \(([^\)]+)\)")

        for world in world_list.worlds:
            for is_dark_world in dark_world_flags(world):
                for area in world.areas:
                    if area.in_dark_aether != is_dark_world:
                        continue
                    for node in area.nodes:
                        if isinstance(node, PickupNode):
                            nodes_by_world[world.correct_name(is_dark_world)].append(node)
                            match = pickup_match.match(node.name)
                            if match is not None:
                                node_name = match.group(1)
                            else:
                                node_name = node.name
                            node_names[node] = f"{world_list.nodes_to_area(node).name} ({node_name})"

        for world_name in sorted(nodes_by_world.keys()):
            spoiler = Foldable(world_name)
            vbox_layout = QtWidgets.QVBoxLayout()
            
            first_node = True
            for node in sorted(nodes_by_world[world_name], key=node_names.get):
                if not first_node:
                    separator_line = QFrame()
                    separator_line.setFrameShape(QFrame.HLine)
                    separator_line.setFrameShadow(QFrame.Sunken)
                    transparent = QGraphicsOpacityEffect(separator_line)
                    transparent.setOpacity(0.33)
                    separator_line.setGraphicsEffect(transparent)
                    vbox_layout.addWidget(separator_line)
                else:
                    first_node = False

                row_widget = LocationPoolRowWidget(node, node_names[node])
                vbox_layout.addWidget(row_widget)
                row_widget.changed.connect(functools.partial(self._on_location_changed, row_widget))
                self._row_widget_for_node[node] = row_widget

            spoiler.set_content_layout(vbox_layout)
            self.locations_scroll_area_layout.addWidget(spoiler)
        
        self.locations_scroll_area_layout.addItem(QSpacerItem(5,5, QSizePolicy.Expanding, QSizePolicy.Expanding))

    @property
    def uses_patches_tab(self) -> bool:
        return False

    @property
    def game_enum(self) -> RandovaniaGame:
        return self.game_description.game

    def on_preset_changed(self, preset: Preset):
        available_locations = preset.configuration.available_locations

        self._during_batch_update = True

        self.check_major_minor.setChecked(available_locations.randomization_mode == RandomizationMode.MAJOR_MINOR_SPLIT)

        for node, row_widget in self._row_widget_for_node.items():
            can_have_progression = node.pickup_index not in available_locations.excluded_indices
            row_widget.set_can_have_progression(can_have_progression)

        self._during_batch_update = False

    def _on_update_randomization_mode(self):
        with self._editor as editor:
            mode = RandomizationMode.MAJOR_MINOR_SPLIT if self.check_major_minor.isChecked() else RandomizationMode.FULL
            editor.available_locations = dataclasses.replace(editor.available_locations, randomization_mode=mode)
            for node, row_widget in self._row_widget_for_node.items():
                if mode == RandomizationMode.MAJOR_MINOR_SPLIT and not node.major_location:
                    row_widget.set_can_have_progression(False)
                    row_widget.setEnabled(False)
                else:
                    row_widget.setEnabled(True)
                    row_widget.set_can_have_progression(True)
                    

    def _on_location_changed(self, row_widget: LocationPoolRowWidget):
        if self._during_batch_update:
            return
        with self._editor as editor:
            editor.available_locations = editor.available_locations.ensure_index(row_widget.node.pickup_index,
                                                                                 not row_widget.can_have_progression)



from dataclasses import dataclass
from typing import Dict, Tuple

from randovania.game_description.item.ammo import Ammo
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.major_item import MajorItem
from randovania.games.game import RandovaniaGame


@dataclass(frozen=True)
class ItemDatabase:
    item_categories: Dict[str, ItemCategory]
    major_items: Dict[str, MajorItem]
    ammo: Dict[str, Ammo]
    default_items: Dict[ItemCategory, Tuple[MajorItem, ...]]


def read_database(database_data: Dict, game: RandovaniaGame) -> ItemDatabase:
    """
    :param database_data:
    :param game:
    :return:
    """
    item_categories = {
        name: ItemCategory.from_json(name, value)
        for name, value in database_data["item_categories"].items()
    }

    major_items = {
        name: MajorItem.from_json(name, value, game, item_categories)
        for name, value in database_data["items"].items()
    }

    ammo = {
        name: Ammo.from_json(name, value, game, item_categories)
        for name, value in database_data["ammo"].items()
    }

    default_items = {
        item_categories[category_name]: tuple(major_items[item_name] for item_name in value)
        for category_name, value in database_data["default_items"].items()
    }

    return ItemDatabase(item_categories, major_items, ammo, default_items)


def write_database(database: ItemDatabase) -> Dict:
    """

    :param database:
    :return:
    """
    item_categories = {
        name: item_category.as_json
        for name, item_category in database.item_categories.items()    
    }

    major_items_data = {
        name: item.as_json
        for name, item in database.major_items.items()
    }

    ammo_data = {
        name: ammo.as_json
        for name, ammo in database.ammo.items()
    }

    default_data = {
        category.name: [item.name for item in items]
        for category, items in database.default_items.items()
    }

    return {"item_categories": item_categories, "items": major_items_data, "ammo": ammo_data, "default_items": default_data}



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'trick_details_popup.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_TrickDetailsPopup(object):
    def setupUi(self, TrickDetailsPopup):
        if not TrickDetailsPopup.objectName():
            TrickDetailsPopup.setObjectName(u"TrickDetailsPopup")
        TrickDetailsPopup.resize(416, 337)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(TrickDetailsPopup.sizePolicy().hasHeightForWidth())
        TrickDetailsPopup.setSizePolicy(sizePolicy)
        TrickDetailsPopup.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(TrickDetailsPopup)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.title_label = QLabel(TrickDetailsPopup)
        self.title_label.setObjectName(u"title_label")
        self.title_label.setWordWrap(True)
        self.title_label.setOpenExternalLinks(True)

        self.verticalLayout.addWidget(self.title_label)

        self.scroll_area = QScrollArea(TrickDetailsPopup)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 396, 191))
        self.verticalLayout_2 = QVBoxLayout(self.scroll_area_contents)
        self.verticalLayout_2.setSpacing(6)
        self.verticalLayout_2.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.area_list_label = QLabel(self.scroll_area_contents)
        self.area_list_label.setObjectName(u"area_list_label")
        self.area_list_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.area_list_label.setWordWrap(True)

        self.verticalLayout_2.addWidget(self.area_list_label)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        self.button_box = QDialogButtonBox(TrickDetailsPopup)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setStandardButtons(QDialogButtonBox.Close)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(TrickDetailsPopup)

        QMetaObject.connectSlotsByName(TrickDetailsPopup)
    # setupUi

    def retranslateUi(self, TrickDetailsPopup):
        TrickDetailsPopup.setWindowTitle(QCoreApplication.translate("TrickDetailsPopup", u"Item Configuration", None))
        self.title_label.setText(QCoreApplication.translate("TrickDetailsPopup", u"<html><head/><body><p><span style=\" font-weight:600;\">{trick.long_name}</span></p><p>{trick.description}</p><p>{trick.long_name} with level {level} can be used in the rooms listed below.</p><p>Click a room name to open it in the Data Visualizer for more details.</p></body></html>", None))
        self.area_list_label.setText(QCoreApplication.translate("TrickDetailsPopup", u"<html><head/><body><p><a href=\"data-editor://World/Area\">INSERT AREAS!</a></p></body></html>", None))
    # retranslateUi




import copy
import json
import os
import time
import logging
from pathlib import Path
from typing import NamedTuple, List

import nod
from retro_data_structures.asset_provider import AssetProvider, InvalidAssetId, UnknownAssetId
from retro_data_structures.conversion import conversions
from retro_data_structures.conversion.asset_converter import AssetConverter
from retro_data_structures.dependencies import all_converted_dependencies
from retro_data_structures.formats import PAK, format_for
from retro_data_structures.game_check import Game

from randovania import get_data_path
from randovania.games.game import RandovaniaGame
from randovania.interface_common.options import Options
from randovania.lib import status_update_lib
from randovania.lib.status_update_lib import ProgressUpdateCallable


def prime_asset_provider() -> AssetProvider:
    opt = Options.with_default_data_dir()
    opt.load_from_disk()
    prime_1_options = opt.options_for_game(RandovaniaGame.METROID_PRIME)
    if prime_1_options.input_path is None:
        raise RuntimeError("No prime1 iso!")

    result = nod.open_disc_from_image(os.fspath(prime_1_options.input_path))
    if result is None:
        raise RuntimeError(f"Invalid prime1 iso: {prime_1_options.input_path}")

    part = result[0].get_data_partition()
    pak_file_names: List[str] = [
        name
        for name in part.files()
        if name.lower().endswith(".pak")
    ]
    pak_files = [
        part.read_file(name)
        for name in pak_file_names
    ]

    return AssetProvider(Game.PRIME, pak_file_names, pak_files)


def echoes_asset_provider() -> AssetProvider:
    internal_copies_path = Options.with_default_data_dir().internal_copies_path
    paks_path = internal_copies_path.joinpath("prime2", "contents", "files")

    all_files = list(paks_path.glob("*.pak"))
    return AssetProvider(Game.ECHOES, all_files)


def create_scan_visor_assets():
    dark_visor_ancs_id = 2233160302
    scan_visor_cmdl = 2949054468

    with echoes_asset_provider() as asset_provider:
        dark_visor_ancs = asset_provider.get_asset(dark_visor_ancs_id)

        scan_visor_ancs = copy.deepcopy(dark_visor_ancs)
        scan_visor_ancs.character_set.characters[0].model_id = scan_visor_cmdl


class Asset(NamedTuple):
    ancs: int
    cmdl: int
    character: int = 0
    scale: float = 1.0


prime1_assets = {
    "Charge Beam": Asset(0xE3CBC3F3, 0xC5472401),
    "Power Beam": Asset(0x00000000, 0x00000000),
    "Wave Beam": Asset(0x09881302, 0x009771B9),
    "Ice Beam": Asset(0x52A3B1A4, 0xDA25B1BE),
    "Plasma Beam": Asset(0x6397CC1B, 0xA792116A),
    # "Missile": Asset(0xA9B8E446, 0x2D7E6590),
    "Grapple Beam": Asset(0xC5B5ED4D, 0xF86621C9),
    "Combat Visor": Asset(0x00000000, 0x00000000),
    "Scan Visor": Asset(0x00000000, 0x00000000),
    "Thermal Visor": Asset(0x9F0C908A, 0x61DAB956),
    "X-Ray Visor": Asset(0x9F0C908A, 0x61DAB956),
    "Space Jump Boots": Asset(0x999E81FE, 0xA10715DA),
    "Energy Tank": Asset(0xF37BCBC7, 0x86908399),
    "Morph Ball": Asset(0x2D0FD5C9, 0x903E8AC5, character=0, scale=1.6),
    "Morph Ball Bomb": Asset(0xDA110E43, 0xB5544D27),
    "Boost Ball": Asset(0x2D0FD5C9, 0x903E8AC5),
    "Spider Ball": Asset(0x2D0FD5C9, 0x79D95DEC, character=2),
    "Power Bomb": Asset(0xF19131AD, 0xD532BDB8, scale=4.668),
    "Power Bomb Expansion": Asset(0x0B5BBF9E, 0x227D7166),
    "Power Suit": Asset(0x00000000, 0x00000000),
    "Varia Suit": Asset(0xA3E787B7, 0xCD995C16),
    "Gravity Suit": Asset(0x27A97006, 0x95946E41),
    "Phazon Suit": Asset(0x00000000, 0x00000000),
    "Super Missile": Asset(0x7C04E388, 0x853A56F0, character=0, scale=2.0),
    "Wavebuster": Asset(0x7C04E388, 0x74A39FE6, character=2, scale=2.0),
    "Ice Spreader": Asset(0x7C04E388, 0x85BA7ACB, character=3, scale=2.0),
    "Flamethrower": Asset(0x7C04E388, 0xC54BBF68, character=1, scale=2.0),
    "Artifact of Truth": Asset(0xFAA9C708, 0x884E88DC, character=3),
    "Artifact of Strength": Asset(0xFAA9C708, 0xFFD05A2C, character=1),
    "Artifact of Elder": Asset(0xFAA9C708, 0x64751643, character=5),
    "Artifact of Wild": Asset(0xFAA9C708, 0x10EDFFCC, character=6),
    "Artifact of Lifegiver": Asset(0xFAA9C708, 0x8B48B3A3, character=7),
    "Artifact of Warrior": Asset(0xFAA9C708, 0xFCD66153, character=8),
    "Artifact of Chozo": Asset(0xFAA9C708, 0x67732D3C, character=9),
    "Artifact of Nature": Asset(0xFAA9C708, 0x15E7B24D, character=10),
    "Artifact of Sun": Asset(0xFAA9C708, 0x8E42FE22, character=11),
    "Artifact of World": Asset(0xFAA9C708, 0x12174A4C, character=0),
    "Artifact of Spirit": Asset(0xFAA9C708, 0x89B20623, character=2),
    "Artifact of Newborn": Asset(0xFAA9C708, 0xFE2CD4D3, character=4),
}


def convert_prime1_pickups(echoes_files_path: Path, randomizer_data: dict, status_update: ProgressUpdateCallable):
    next_id = 0xFFFF0000

    def id_generator(asset_type):
        nonlocal next_id
        result = next_id
        while asset_provider.asset_id_exists(result):
            result += 1

        next_id = result + 1
        return result

    try:
        asset_provider = prime_asset_provider()
    except RuntimeError:
        return

    updaters = status_update_lib.split_progress_update(status_update, 3)

    start = time.time()
    with asset_provider:
        updaters[0]("Loading Prime 1 PAKs", 0)
        converter = AssetConverter(
            target_game=Game.ECHOES,
            asset_providers={Game.PRIME: asset_provider},
            id_generator=id_generator,
            converters=conversions.converter_for,
        )
        updaters[0]("Finished loading Prime 1 PAKs", 1)
        # logging.debug(f"Finished loading PAKs: {time.time() - start}")

        result = {}
        num_assets = len(prime1_assets)
        for i, (name, asset) in enumerate(prime1_assets.items()):
            updaters[1](f"Converting {name} from Prime 1", i / num_assets)
            if asset.ancs != 0 and asset.cmdl != 0:
                result[name] = Asset(
                    ancs=converter.convert_id(asset.ancs, Game.PRIME),
                    cmdl=converter.convert_id(asset.cmdl, Game.PRIME),
                    character=asset.character,
                    scale=asset.scale,
                )
        updaters[1]("Finished converting Prime 1 assets", 1)
    end = time.time()
    # logging.debug(f"Time took: {end - start}")

    pak_updaters = status_update_lib.split_progress_update(updaters[2], 5)
    for pak_i in range(1, 6):
        pak_status = pak_updaters[pak_i - 1]
        pak_path = echoes_files_path.joinpath("files", f"Metroid{pak_i}.pak")
        pak_status(f"Preparing to write custom assets to Echoes {pak_path.name}", 0)
        num_assets = len(converter.converted_assets) + 1

        new_pak = PAK.parse(
            pak_path.read_bytes(),
            target_game=Game.ECHOES,
        )
        for i, new_asset in enumerate(converter.converted_assets.values()):
            if new_asset.type.upper() == "EVNT":
                continue

            pak_status(f"Writing new asset {new_asset.type} 0x{new_asset.id:08X} to {pak_path.name}", i / num_assets)
            new_pak.resources.append({
                "compressed": 0,
                "asset": {
                    "type": new_asset.type,
                    "id": new_asset.id,
                },
                "contents": {
                    "value": format_for(new_asset.type).build(new_asset.resource, target_game=Game.ECHOES),
                },
            })

        pak_status(f"Writing new {pak_path.name}", (num_assets - 1) / num_assets)
        PAK.build_file(new_pak, pak_path, target_game=Game.ECHOES)

    converted_dependencies = all_converted_dependencies(converter)

    # logging.debug("Updating RandomizerData.json")
    start = time.time()
    for name, asset in result.items():
        dependencies = [
            {"AssetID": dep.id, "Type": dep.type.upper()}
            for dep in converted_dependencies[asset.ancs] | converted_dependencies[asset.cmdl]
        ]
        randomizer_data["ModelData"].append({
            "Index": len(randomizer_data["ModelData"]),
            "Name": f"prime1_{name}",
            "Model": asset.cmdl,
            "ScanModel": 0xFFFFFFFF,
            "AnimSet": asset.ancs,
            "Character": asset.character,
            "DefaultAnim": 0,
            "Rotation": [0.0, 0.0, 0.0],
            "Scale": [asset.scale, asset.scale, asset.scale],
            "OrbitOffset": [0.0, 0.0, 0.0],
            "Lighting": {
                "CastShadow": True,
                "UnknownBool1": True,
                "UseWorldLighting": 1,
                "UnknownBool2": False
            },
            "Assets": dependencies
        })

    end = time.time()
    # logging.debug(f"Time took: {end - start}")


def convert_prime2_pickups():
    next_id = 0xFFFF0000

    randomizer_data_path = get_data_path().joinpath("ClarisPrimeRandomizer", "RandomizerData.json")
    with randomizer_data_path.open() as randomizer_data_file:
        randomizer_data = json.load(randomizer_data_file)

    def id_generator(asset_type):
        nonlocal next_id
        result = next_id
        while asset_provider.asset_id_exists(result):
            result += 1

        next_id = result + 1
        return result

    start = time.time()
    with echoes_asset_provider() as asset_provider:
        logging.info("Loading PAKs")
        converter = AssetConverter(
            target_game=Game.PRIME,
            asset_providers={Game.ECHOES: asset_provider},
            id_generator=id_generator,
            converters=conversions.converter_for,
        )
        logging.info(f"Finished loading PAKs: {time.time() - start}")

        result = {}
        for data in randomizer_data["ModelData"]:
            if data["Model"] != Game.ECHOES.invalid_asset_id and data["AnimSet"] != Game.ECHOES.invalid_asset_id:
                try:
                    result[data["Name"]] = Asset(
                        ancs=converter.convert_id(data["AnimSet"], Game.ECHOES, missing_assets_as_invalid=False),
                        cmdl=converter.convert_id(data["Model"], Game.ECHOES, missing_assets_as_invalid=False),
                        character=data["Character"],
                        scale=data["Scale"][0],
                    )
                except (InvalidAssetId, UnknownAssetId) as e:
                    logging.error("Unable to convert {}: {}".format(data["Name"], e))
    end = time.time()
    logging.info(f"Time took: {end - start}")

    start = time.time()
    converted_dependencies = all_converted_dependencies(converter)

    new_id_to_old = {
        new_id: old_id
        for (_, old_id), new_id in converter.converted_ids.items()
    }

    Path("converted").mkdir(exist_ok=True)
    with open("converted/meta.json", "w") as meta_out:
        json.dump({
            "items": {
                name: {
                    "ancs": asset.ancs,
                    "cmdl": asset.cmdl,
                    "character": asset.character,
                    "scale": asset.scale,
                }
                for name, asset in result.items()
            },
            "new_assets": [
                {
                    "old_id": new_id_to_old.get(asset.id),
                    "new_id": asset.id,
                    "type": asset.type,
                    "dependencies": [
                        {"type": dep.type, "id": dep.id}
                        for dep in converted_dependencies[asset.id]
                    ]
                }
                for asset in converter.converted_assets.values()
            ],
        }, meta_out, indent=4)

    for asset in converter.converted_assets.values():
        Path("converted").joinpath(f"{asset.id}.{asset.type.upper()}").write_bytes(
            format_for(asset.type).build(asset.resource, target_game=Game.PRIME)
        )

    logging.info(f"Time took: {time.time() - start}")


if __name__ == '__main__':
    convert_prime2_pickups()



import dataclasses
import struct
from enum import Enum
from typing import List

from randovania.dol_patching.assembler import custom_ppc
from randovania.dol_patching.assembler.ppc import *
from randovania.dol_patching.dol_file import DolFile
from randovania.game_description import default_database
from randovania.games.game import RandovaniaGame
from randovania.games.prime.all_prime_dol_patches import (
    BasePrimeDolVersion, HealthCapacityAddresses,
    DangerousEnergyTankAddresses
)
from randovania.layout.prime2.beam_configuration import BeamConfiguration
from randovania.layout.prime2.echoes_user_preferences import EchoesUserPreferences


@dataclasses.dataclass(frozen=True)
class BeamCostAddresses:
    uncharged_cost: int
    charged_cost: int
    charge_combo_ammo_cost: int
    charge_combo_missile_cost: int
    get_beam_ammo_type_and_costs: int
    is_out_of_ammo_to_shoot: int
    gun_get_player: int
    get_item_amount: int


@dataclasses.dataclass(frozen=True)
class SafeZoneAddresses:
    heal_per_frame_constant: int
    increment_health_fmr: int


@dataclasses.dataclass(frozen=True)
class StartingBeamVisorAddresses:
    player_state_constructor_clean: int
    player_state_constructor_decode: int
    health_info_constructor: int
    enter_morph_ball_state: int
    start_transition_to_visor: int
    reset_visor: int


_PREFERENCES_ORDER = (
    "sound_mode",
    "screen_brightness",
    "screen_x_offset",
    "screen_y_offset",
    "screen_stretch",
    "sfx_volume",
    "music_volume",
    "hud_alpha",
    "helmet_alpha",
)
_FLAGS_ORDER = (
    "hud_lag",
    "invert_y_axis",
    "rumble",
    None,  # crashes, maybe swapBeamsControls
    "hint_system",
    None,  # 5: doesn't crash, unknown
    None,  # 6: doesn't crash, unknown
    None,  # 7: doesn't crash, unknown
)


def apply_game_options_patch(game_options_constructor_offset: int, user_preferences: EchoesUserPreferences,
                             dol_file: DolFile):
    patch = [
        # Unknown purpose, but keep for safety
        stw(r31, 0x1c, r1),
        or_(r31, r3, r3),

        # For a later function call we don't touch
        addi(r3, r1, 0x8),
    ]

    for i, preference_name in enumerate(_PREFERENCES_ORDER):
        value = getattr(user_preferences, preference_name)
        if isinstance(value, Enum):
            value = value.value
        patch.extend([
            li(r0, value),
            stw(r0, (0x04 * i), r31),
        ])

    flag_values = [
        getattr(user_preferences, flag_name)
        if flag_name is not None else False
        for flag_name in _FLAGS_ORDER
    ]
    bit_mask = int("".join(str(int(flag)) for flag in flag_values), 2)
    patch.extend([
        li(r0, bit_mask),
        stb(r0, 0x04 * len(_PREFERENCES_ORDER), r31),
        li(r0, 0),
        stw(r0, 0x2c, r31),
        stw(r0, 0x30, r31),
        stw(r0, 0x34, r31),
    ])

    instructions_space = 34
    instructions_to_fill = instructions_space - len(patch)

    if instructions_to_fill < 0:
        raise RuntimeError(f"Our patch ({len(patch)}) is bigger than the space we have ({instructions_space}).")

    for i in range(instructions_to_fill):
        patch.append(nop())
    dol_file.write_instructions(game_options_constructor_offset + 8 * 4, patch)


def _is_out_of_ammo_patch(symbols: Dict[str, int], ammo_types: List[Tuple[int, int]]):
    def get_beam_ammo_amount(index: int):
        label = f"_after_get_ammo_type_{index}"

        body = [
            lwz(r5, 0x774, r30),  # r5 = get current beam
            addi(r5, r5, 0x1),  # current_beam += 1
            mtspr(CTR, r5),  # count_register = current_beam
        ]

        for beam_index, beam_ammo_types in enumerate(ammo_types):
            instructions = []
            if beam_index + 1 < len(ammo_types):
                instructions.append(bdnz(f"_before_get_ammo_type_{beam_index + 1}_{index}"))

            if beam_ammo_types[index] == -1:
                instructions.extend([
                    li(r3, 0),  # No ammo type, so load result
                    b("_end"),  # and return
                ])
            else:
                instructions.extend([
                    li(r4, beam_ammo_types[index]),
                    b(label),
                ])

            instructions[0].with_label(f"_before_get_ammo_type_{beam_index}_{index}")
            body.extend(instructions)

        body.extend([
            or_(r3, r31, r31).with_label(label),  # arg1 = playerState, arg2 is already there
            li(r5, 1),  # arg3 = true, allow multiplayer ammo stuff
            bl("CPlayerState::GetItemAmount"),  # r3 = ammoCount
        ])

        return body

    get_uncharged_cost = [
        custom_ppc.load_unsigned_32bit(r4, symbols["BeamIdToChargedShotAmmoCost"]),
        lwz(r5, 0x774, r30),  # r5 = get current beam
        rlwinm(r5, r5, 0x2, 0x0, 0x1d),  # r5 *= 4
        lwzx(r4, r4, r5),  # ammoCost_r4 = UnchargedCosts_r4[currentBeam]
    ]
    compare_count_to_cost = [
        cmpw(0, r3, r4),
        bge(3 * 4, relative=True),  # if ammoCount_3 >= ammoCost_r4, goto
        li(r3, 1),  # Not enough ammo, load true
        b("_end"),  # and return
    ]

    return [
        # Save stack
        stwu(r1, -0x10, r1),
        mfspr(r0, LR),
        stw(r0, 0x14, r1),

        # Save r31 and r30
        stw(r31, 0xc, r1),
        stw(r30, 0x8, r1),

        # Save a pointer to CPlayerGun
        or_(r30, r3, r3),
        bl("CPlayerGun::GetPlayer"),

        # Get and save a pointer to CPlayerState
        lwz(r31, 0x1314, r3),

        # check ammo type 1
        *get_beam_ammo_amount(0),  # r3 = ammo amount
        *get_uncharged_cost,  # r4 = uncharged_cost
        *compare_count_to_cost,

        # check ammo type 2
        *get_beam_ammo_amount(1),  # r3 = ammo amount
        *get_uncharged_cost,  # r4 = uncharged_cost
        *compare_count_to_cost,

        # All ammo types for this beam are fine!
        li(r3, 0),
        b("_end"),  # and return

        # end
        lwz(r0, 0x14, r1).with_label("_end"),
        lwz(r31, 0xc, r1),
        lwz(r30, 0x8, r1),
        mtspr(LR, r0),
        addi(r1, r1, 0x10),
        blr(),
    ]


def apply_beam_cost_patch(patch_addresses: BeamCostAddresses, beam_configuration: BeamConfiguration,
                          dol_file: DolFile):
    uncharged_costs = []
    charged_costs = []
    combo_costs = []
    missile_costs = []
    ammo_types = []

    for beam_config in beam_configuration.all_beams:
        uncharged_costs.append(beam_config.uncharged_cost)
        charged_costs.append(beam_config.charged_cost)
        combo_costs.append(beam_config.combo_ammo_cost)
        missile_costs.append(beam_config.combo_missile_cost)
        ammo_types.append((
            beam_config.ammo_a,
            beam_config.ammo_b,
        ))

    # The following patch also changes the fact that the game doesn't check if there's enough ammo for Power Beam
    # we start our patch right after the `addi r3,r31,0x0`
    ammo_type_patch_offset = 0x40
    offset_to_body_end = 0xB4
    ammo_type_patch = [
        lwz(r10, 0x774, r25),  # r10 = get current beam
        rlwinm(r10, r10, 0x2, 0x0, 0x1d),  # r10 *= 4

        lwzx(r0, r3, r10),  # r0 = BeamIdToUnchargedShotAmmoCost[currentBeam]
        stw(r0, 0x0, r29),  # *outBeamAmmoCost = r0

        lwz(r10, 0x774, r25),  # r10 = get current beam
        addi(r10, r10, 0x1),  # r10 = r10 + 1
        mtspr(CTR, r10),  # count_register = r10

        # Power Beam
        bdnz("dark_beam"),  # if (--count_register > 0) goto
        li(r3, ammo_types[0][0]),
        li(r9, ammo_types[0][1]),
        b("update_out_beam_type"),

        # Dark Beam
        bdnz("light_beam").with_label("dark_beam"),  # if (--count_register > 0) goto
        li(r3, ammo_types[1][0]),
        li(r9, ammo_types[1][1]),
        b("update_out_beam_type"),

        # Light Beam
        bdnz("annihilator_beam").with_label("light_beam"),  # if (--count_register > 0) goto
        li(r3, ammo_types[2][0]),
        li(r9, ammo_types[2][1]),
        b("update_out_beam_type"),

        # Annihilator Beam
        li(r3, ammo_types[3][0]).with_label("annihilator_beam"),
        li(r9, ammo_types[3][1]),

        # update_out_beam_type
        stw(r3, 0x0, r27).with_label("update_out_beam_type"),  # *outBeamAmmoTypeA = r3
        stw(r9, 0x0, r28),  # *outBeamAmmoTypeB = r9

        b(patch_addresses.get_beam_ammo_type_and_costs + offset_to_body_end),
        # jump to the code for getting the charged/combo costs and then check if has ammo
        # The address in question is at 0x801ccd64 for NTSC
    ]

    # FIXME: depend on version
    dol_file.symbols["BeamIdToChargedShotAmmoCost"] = patch_addresses.uncharged_cost
    dol_file.symbols["BeamIdToUnchargedShotAmmoCost"] = patch_addresses.charged_cost
    dol_file.symbols["BeamIdToChargeComboAmmoCost"] = patch_addresses.charge_combo_ammo_cost
    dol_file.symbols["g_ChargeComboMissileCosts"] = patch_addresses.charge_combo_missile_cost
    dol_file.symbols["CPlayerGun::IsOutOfAmmoToShoot"] = patch_addresses.is_out_of_ammo_to_shoot
    dol_file.symbols["CPlayerGun::GetPlayer"] = patch_addresses.gun_get_player
    dol_file.symbols["CPlayerState::GetItemAmount"] = patch_addresses.get_item_amount

    uncharged_costs_patch = struct.pack(">llll", *uncharged_costs)
    charged_costs_patch = struct.pack(">llll", *charged_costs)
    combo_costs_patch = struct.pack(">llll", *combo_costs)
    missile_costs_patch = struct.pack(">llll", *missile_costs)

    dol_file.write("BeamIdToChargedShotAmmoCost", uncharged_costs_patch)
    dol_file.write("BeamIdToUnchargedShotAmmoCost", charged_costs_patch)
    dol_file.write("BeamIdToChargeComboAmmoCost", combo_costs_patch)
    dol_file.write("g_ChargeComboMissileCosts", missile_costs_patch)
    dol_file.write_instructions(patch_addresses.get_beam_ammo_type_and_costs + ammo_type_patch_offset,
                                ammo_type_patch)
    dol_file.write_instructions("CPlayerGun::IsOutOfAmmoToShoot", _is_out_of_ammo_patch(dol_file.symbols, ammo_types))


def apply_safe_zone_heal_patch(patch_addresses: SafeZoneAddresses,
                               sda2_base: int,
                               heal_per_second: float,
                               dol_file: DolFile):
    offset = patch_addresses.heal_per_frame_constant - sda2_base

    dol_file.write(patch_addresses.heal_per_frame_constant, struct.pack(">f", heal_per_second / 60))
    dol_file.write_instructions(patch_addresses.increment_health_fmr, [lfs(f1, offset, r2)])


def apply_starting_visor_patch(addresses: StartingBeamVisorAddresses, default_items: dict, dol_file: DolFile):
    visor_order = ["Combat Visor", "Echo Visor", "Scan Visor", "Dark Visor"]
    beam_order = ["Power Beam", "Dark Beam", "Light Beam", "Annihilator Beam"]

    default_visor = visor_order.index(default_items["visor"])
    default_beam = beam_order.index(default_items["beam"])

    # Patch CPlayerState constructor with default values
    dol_file.write_instructions(addresses.player_state_constructor_clean + 0x54, [
        bl(addresses.health_info_constructor),

        li(r0, default_beam),
        stw(r0, 0xc, r30),  # xc_currentBeam

        li(r0, default_visor),
        stw(r0, 0x30, r30),  # x30_currentVisor
        stw(r0, 0x34, r30),  # x34_transitioningVisor

        li(r3, 0),
    ])

    # Patch CPlayerState constructor for loading save files
    dol_file.write_instructions(addresses.player_state_constructor_decode + 0x5C, [
        li(r0, default_visor),
        stw(r0, 0x30, r30),
        stw(r0, 0x34, r30),
    ])

    # Patch EnterMorphBallState's call for StartTransitionToVisor to use the new default visor
    dol_file.write_instructions(addresses.enter_morph_ball_state + 0xE8, [
        li(r4, default_visor),
    ])

    # Patch CPlayerState::ResetVisor so elevators use the new default visor
    dol_file.write_instructions(addresses.reset_visor, [
        li(r0, default_visor),
    ])


@dataclasses.dataclass(frozen=True)
class EchoesDolVersion(BasePrimeDolVersion):
    health_capacity: HealthCapacityAddresses
    dangerous_energy_tank: DangerousEnergyTankAddresses
    game_options_constructor_address: int
    beam_cost_addresses: BeamCostAddresses
    safe_zone: SafeZoneAddresses
    starting_beam_visor: StartingBeamVisorAddresses
    anything_set_address: int
    rs_debugger_printf_loop_address: int
    unvisited_room_names_address: int
    cworldtransmanager_sfxstart: int
    powerup_should_persist: int


def apply_fixes(version: EchoesDolVersion, dol_file: DolFile):
    resource_database = default_database.game_description_for(RandovaniaGame.METROID_PRIME_ECHOES).resource_database

    dol_file.symbols["CMapWorldInfo::IsAnythingSet"] = version.anything_set_address

    dol_file.write_instructions("CMapWorldInfo::IsAnythingSet", [
        li(r3, 1),
        blr(),
    ])

    dol_file.write_instructions(version.rs_debugger_printf_loop_address, [
        nop(),
    ])

    for item in ["Double Damage", "Unlimited Missiles", "Unlimited Beam Ammo"]:
        index = resource_database.get_item_by_name(item).index
        dol_file.write(version.powerup_should_persist + index, b"\x01")


def apply_unvisited_room_names(version: EchoesDolVersion, dol_file: DolFile, enabled: bool):
    # In CAutoMapper::Update, the function checks for `mwInfo.IsMapped` then `mwInfo.IsAreaVisited` and if both are
    # false, sets a variable to false. This variable indicates if the room name is displayed used.
    dol_file.write_instructions(version.unvisited_room_names_address, [
        li(r28, 1 if enabled else 0),
    ])


def apply_teleporter_sounds(version: EchoesDolVersion, dol_file: DolFile, enabled: bool):
    dol_file.symbols["CWorldTransManager::SfxStart"] = version.cworldtransmanager_sfxstart

    if enabled:
        inst = stwu(r1, -0x20, r1)
    else:
        inst = blr()

    dol_file.write_instructions("CWorldTransManager::SfxStart", [
        inst
    ])


def freeze_player():
    return [
        lfs(f1, -0x707c, r2),  # timeout = 5.0f
        lwz(r3, 0x14fc, r31),  # player = manager->players[0]
        lhz(r6, -0x40da, r2),  # sfxId = kInvalidSoundId
        or_(r4, r31, r31),  # mgr
        li(r5, -0x1),  # steamTextureId
        li(r7, -0x1),  # iceTextureId
        bl("CPlayer::Freeze"),
    ]



import argparse
import collections
import json
import re
import typing
from pathlib import Path

from randovania.layout.layout_description import LayoutDescription


def iterate_with_log(x):
    try:
        import tqdm
        return tqdm.tqdm(x)
    except ImportError:
        print("WARNING: tqdm not found. Use `pip install tqdm` to have progress feedback.")
        return x


def read_json(path: Path) -> dict:
    with path.open() as x:
        return json.load(x)


_KEY_MATCH = re.compile(r"Key (\d+)")
_PLAYER_MATCH = re.compile(r" for Player \d+")


def _filter_item_name(name: str) -> str:
    return _PLAYER_MATCH.sub("", _KEY_MATCH.sub("Key", name))


def create_report(seeds_dir: str, output_file: str):
    def item_creator():
        return collections.defaultdict(list)

    item_name_to_location = collections.defaultdict(item_creator)

    seed_files = list(Path(seeds_dir).glob(f"**/*.{LayoutDescription.file_extension()}"))
    seed_files.extend(Path(seeds_dir).glob("**/*.json"))
    for seed in iterate_with_log(seed_files):
        seed = typing.cast(Path, seed)
        try:
            seed_data = read_json(seed)
        except json.JSONDecodeError:
            continue

        for item_order in seed_data["item_order"]:
            if " as " in item_order:
                continue

            item_name, item_location = item_order.split(" at ", 1)
            item_name = _filter_item_name(item_name)
            item_location = item_location.split(" with hint ", 1)[0]

            if "Expansion" in item_name or item_name == "Energy Tank":
                continue

            item_name_to_location[item_name][item_location].append(str(seed.relative_to(seeds_dir)))

    final_results = {
        item_name: {
            location: seeds
            for location, seeds in sorted(locations.items(), key=lambda it: len(it[1]))
            if len(seeds) < 250
        }
        for item_name, locations in sorted(item_name_to_location.items(), key=lambda it: it[0])
    }

    with open(output_file, "w") as output:
        json.dump(final_results, output, indent=4, separators=(',', ': '))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("seeds_dir")
    parser.add_argument("output_file")
    args = parser.parse_args()
    create_report(args.seeds_dir, args.output_file)


if __name__ == "__main__":
    main()



from typing import Iterator, Dict, Tuple

from randovania.game_description.game_description import GameDescription
from randovania.game_description.requirements import RequirementSet
from randovania.game_description.world.node import Node, ResourceNode
from randovania.resolver.state import State


class GeneratorReach:
    @classmethod
    def reach_from_state(cls,
                         game: GameDescription,
                         initial_state: State,
                         ) -> "GeneratorReach":
        raise NotImplementedError()

    # Game related methods

    @property
    def game(self) -> GameDescription:
        raise NotImplementedError()

    def victory_condition_satisfied(self):
        return self.game.victory_condition.satisfied(self.state.resources, self.state.energy,
                                                     self.state.resource_database)

    @property
    def all_nodes(self) -> Tuple[Node, ...]:
        return self.game.world_list.all_nodes

    # ASDF

    @property
    def state(self) -> State:
        raise NotImplementedError()

    def advance_to(self, new_state: State,
                   is_safe: bool = False,
                   ) -> None:
        raise NotImplementedError()

    def act_on(self, node: ResourceNode) -> None:
        raise NotImplementedError()

    # Node stuff

    def is_reachable_node(self, node: Node) -> bool:
        raise NotImplementedError()

    @property
    def connected_nodes(self) -> Iterator[Node]:
        """
        An iterator of all nodes there's an path from the reach's starting point. Similar to is_reachable_node
        :return:
        """
        raise NotImplementedError()

    @property
    def nodes(self) -> Iterator[Node]:
        raise NotImplementedError()

    @property
    def safe_nodes(self) -> Iterator[Node]:
        raise NotImplementedError()

    def is_safe_node(self, node: Node) -> bool:
        raise NotImplementedError()

    def unreachable_nodes_with_requirements(self) -> Dict[Node, RequirementSet]:
        raise NotImplementedError()






from PySide2 import QtWidgets, QtCore

from randovania.game_description import default_database
from randovania.game_description.world.node import LogbookNode, LoreType
from randovania.games.game import RandovaniaGame
from randovania.generator.item_pool import pickup_creator


def prime1_hint_text():
    db = default_database.resource_database_for(RandovaniaGame.METROID_PRIME)
    artifact = pickup_creator.create_artifact(0, 0, db)

    result = [(
        "Artifact",
        artifact.item_category,
        artifact.broad_category,
    )]
    return result


def prime2_hint_text():
    db = default_database.resource_database_for(RandovaniaGame.METROID_PRIME_ECHOES)

    result = []

    for temple in range(3):
        key = pickup_creator.create_dark_temple_key(0, temple, db)
        result.append((
            key.name.replace(" 1", "").strip(),
            key.item_category,
            key.broad_category,
        ))

    key = pickup_creator.create_sky_temple_key(0, db)
    result.append((
        "Sky Temple Key",
        key.item_category,
        key.broad_category,
    ))

    return result


def prime3_hint_text():
    db = default_database.resource_database_for(RandovaniaGame.METROID_PRIME_CORRUPTION)
    cell = pickup_creator.create_energy_cell(0, db)

    result = [(
        "Energy Cell",
        cell.item_category,
        cell.broad_category,
    )]
    return result


_GAME_SPECIFIC = {
    RandovaniaGame.METROID_PRIME: prime1_hint_text,
    RandovaniaGame.METROID_PRIME_ECHOES: prime2_hint_text,
    RandovaniaGame.METROID_PRIME_CORRUPTION: prime3_hint_text,
}


def update_hints_text(game: RandovaniaGame,
                      hint_item_names_tree_widget: QtWidgets.QTableWidget,
                      ):
    item_database = default_database.item_database_for_game(game)

    rows = []

    for item in item_database.major_items.values():
        rows.append((
            item.name,
            item.item_category.hint_details[1],
            item.item_category.general_details[1],
            item.broad_category.hint_details[1],
        ))

    for name, item_category, broad_category in _GAME_SPECIFIC[game]():
        rows.append((
            name,
            item_category.hint_details[1],
            item_category.general_details[1],
            broad_category.hint_details[1],
        ))

    for ammo in item_database.ammo.values():
        rows.append((
            ammo.name,
            ammo.item_category.hint_details[1],
            ammo.item_category.general_details[1],
            ammo.broad_category.hint_details[1],
        ))

    hint_item_names_tree_widget.setSortingEnabled(False)
    hint_item_names_tree_widget.setRowCount(len(rows))
    for i, elements in enumerate(rows):
        for j, element in enumerate(elements):
            hint_item_names_tree_widget.setItem(i, j, QtWidgets.QTableWidgetItem(element))

    for i in range(4):
        hint_item_names_tree_widget.resizeColumnToContents(i)

    hint_item_names_tree_widget.setSortingEnabled(True)


def update_hint_locations(game: RandovaniaGame,
                          hint_tree_widget: QtWidgets.QTreeWidget):
    game_description = default_database.game_description_for(game)

    number_for_hint_type = {
        hint_type: i + 1
        for i, hint_type in enumerate(LoreType)
    }
    used_hint_types = set()

    hint_tree_widget.clear()
    hint_tree_widget.setSortingEnabled(False)

    # TODO: This ignores the Dark World names. But there's currently no logbook nodes in Dark World.
    for world in game_description.world_list.worlds:

        world_item = QtWidgets.QTreeWidgetItem(hint_tree_widget)
        world_item.setText(0, world.name)
        world_item.setExpanded(True)

        for area in world.areas:
            hint_types = {}

            for node in area.nodes:
                if isinstance(node, LogbookNode):
                    if node.required_translator is not None:
                        hint_types[node.lore_type] = node.required_translator.short_name
                    else:
                        hint_types[node.lore_type] = "✓"

            if hint_types:
                area_item = QtWidgets.QTreeWidgetItem(world_item)
                area_item.setText(0, area.name)

                for hint_type, text in hint_types.items():
                    area_item.setText(number_for_hint_type[hint_type], text)
                    used_hint_types.add(hint_type)

    hint_tree_widget.resizeColumnToContents(0)
    hint_tree_widget.setSortingEnabled(True)
    hint_tree_widget.sortByColumn(0, QtCore.Qt.AscendingOrder)

    for hint_type in used_hint_types:
        hint_tree_widget.headerItem().setText(number_for_hint_type[hint_type], hint_type.long_name)



from typing import List

from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo, CurrentResources
from randovania.generator.item_pool.pool_creator import calculate_pool_results
from randovania.layout.base.base_configuration import BaseConfiguration

_RESOURCE_NAME_TRANSLATION = {
    'Temporary Missile': 'Missile',
    'Temporary Power Bomb': 'Power Bomb',
}
_ITEMS_TO_PLURALIZE = {
    "Missile",
    "Power Bomb",
}


def add_quantity_to_resource(resource: str, quantity: int, always_add_quantity: bool = False) -> str:
    if always_add_quantity or quantity > 1:
        first_part = f"{quantity} "
    else:
        first_part = ""
    return "{}{}{}".format(first_part, resource, "s" if quantity > 1 and resource in _ITEMS_TO_PLURALIZE else "")


def resource_user_friendly_name(resource: ResourceInfo) -> str:
    """
    Gets a name that we should display to the user for the given resource.
    :param resource:
    :return:
    """
    return _RESOURCE_NAME_TRANSLATION.get(resource.long_name, resource.long_name)


def additional_starting_items(layout_configuration: BaseConfiguration,
                              resource_database: ResourceDatabase,
                              starting_items: CurrentResources) -> List[str]:
    initial_items = calculate_pool_results(layout_configuration, resource_database)[2]

    return [
        add_quantity_to_resource(resource_user_friendly_name(item), quantity)
        for item, quantity in sorted(starting_items.items(), key=lambda a: resource_user_friendly_name(a[0]))
        if 0 < quantity != initial_items.get(item, 0)
    ]



import functools
import json
from pathlib import Path
from typing import Tuple

from randovania import get_data_path
from randovania.game_description import data_reader
from randovania.games.game import RandovaniaGame
from randovania.games.binary_data import decode_file_path


@functools.lru_cache()
def read_json_then_binary(game: RandovaniaGame) -> Tuple[Path, dict]:
    dir_path = get_data_path().joinpath("json_data", f"{game.value}")
    if dir_path.exists():
        return dir_path, data_reader.read_split_file(dir_path)

    json_path = get_data_path().joinpath("json_data", f"{game.value}.json")
    if json_path.exists():
        with json_path.open("r") as open_file:
            return json_path, json.load(open_file)

    binary_path = get_data_path().joinpath("binary_data", f"{game.value}.bin")
    return binary_path, decode_file_path(binary_path)



import dataclasses
from pathlib import Path
from typing import Optional

from PySide2.QtWidgets import QMessageBox, QDialog, QLabel, QRadioButton

from randovania.games.game import RandovaniaGame
from randovania.games.patcher import Patcher
from randovania.gui.generated.game_input_dialog_ui import Ui_GameInputDialog
from randovania.gui.lib import common_qt_lib
from randovania.interface_common.options import Options

_VALID_GAME_TEXT = "(internal game copy)"


class GameInputDialog(QDialog, Ui_GameInputDialog):
    _options: Options
    _has_spoiler: bool
    _game: RandovaniaGame
    _prompt_input_file: bool
    _current_lock_state: bool = True
    _selected_output_format: str

    def __init__(self, options: Options, patcher: Patcher, word_hash: str, spoiler: bool, game: RandovaniaGame):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self._options = options
        self._has_spoiler = spoiler
        self._game = game

        self.patcher = patcher
        self.default_output_name = patcher.default_output_file(word_hash)
        self.check_extracted_game()

        per_game = options.options_for_game(self._game)

        description_text = "<html><head/><body><p>In order to create the randomized game, an ISO file of {} for the Nintendo Gamecube is necessary.</p>".format(
            game.long_name)
        if not patcher.uses_input_file_directly:
            description_text += "<p>After using it once, a copy is kept by Randovania for later use.</p>"
        description_text += "</body></html>"
        self.description_label.setText(description_text)

        # Input
        self.input_file_edit.textChanged.connect(self._validate_input_file)
        self.input_file_button.clicked.connect(self._on_input_file_button)

        # Output
        self.output_file_edit.textChanged.connect(self._validate_output_file)
        self.output_file_button.clicked.connect(self._on_output_file_button)

        # Output format
        if per_game.output_format is not None:
            self._selected_output_format = per_game.output_format
        else:
            self._selected_output_format = patcher.valid_output_file_types[0]
        if len(patcher.valid_output_file_types) > 1:
            layout = self.output_format_layout
            output_label = QLabel()
            output_label.setText("Output Format")
            layout.addWidget(output_label)
            for filetype in patcher.valid_output_file_types:
                radio = QRadioButton("." + filetype, self)
                if filetype == self._selected_output_format:
                    radio.setChecked(True)
                radio.toggled.connect(self._on_output_format_changed)
                layout.addWidget(radio)

        # Spoiler
        self.auto_save_spoiler_check.setEnabled(spoiler)
        self.auto_save_spoiler_check.setChecked(options.auto_save_spoiler)

        # Accept/Reject
        self.accept_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)

        self.input_file_edit.has_error = False
        self.output_file_edit.has_error = False

        if self._prompt_input_file and per_game.input_path is not None:
            self.input_file_edit.setText(str(per_game.input_path))

        if per_game.output_directory is not None:
            self.output_file_edit.setText(str(per_game.output_directory.joinpath("{}.{}".format(self.default_output_name, self._selected_output_format))))

        self._validate_input_file()
        self._validate_output_file()

    def save_options(self):
        with self._options as options:
            if self._has_spoiler:
                options.auto_save_spoiler = self.auto_save_spoiler

            per_game = options.options_for_game(self._game)
            per_game_changes = {"output_directory": self.output_file.parent}
            per_game_changes["output_format"] = self._selected_output_format
            if self._prompt_input_file:
                per_game_changes["input_path"] = self.input_file

            options.set_options_for_game(self._game, dataclasses.replace(per_game, **per_game_changes))

    # Getters
    @property
    def input_file(self) -> Optional[Path]:
        if self._prompt_input_file:
            return Path(self.input_file_edit.text())

    @property
    def output_file(self) -> Path:
        return Path(self.output_file_edit.text())

    @property
    def auto_save_spoiler(self) -> bool:
        return self.auto_save_spoiler_check.isChecked()

    def _update_accept_button(self):
        self.accept_button.setEnabled(not (self.input_file_edit.has_error or self.output_file_edit.has_error))

    # Checks
    def check_extracted_game(self):
        self._prompt_input_file = (self.patcher.uses_input_file_directly or
                                   not self.patcher.has_internal_copy(self._options.internal_copies_path))
        self.input_file_edit.setEnabled(self._prompt_input_file)

        if self._prompt_input_file:
            self.input_file_button.setText("Select File")
        else:
            self.input_file_button.setText("Delete internal copy")
            self.input_file_edit.setText(_VALID_GAME_TEXT)

    # Input file
    def _validate_input_file(self):
        if self._prompt_input_file:
            has_error = not self.input_file.is_file()
        else:
            has_error = self.input_file_edit.text() != _VALID_GAME_TEXT

        common_qt_lib.set_error_border_stylesheet(self.input_file_edit, has_error)
        self._update_accept_button()

    def _on_input_file_button(self):
        if self._prompt_input_file:
            existing_file = None
            if self.input_file.is_file():
                existing_file = self.input_file
            elif self.input_file_edit.text() and self.input_file.parent.is_dir():
                existing_file = self.input_file.parent

            input_file = common_qt_lib.prompt_user_for_vanilla_input_file(self, self.patcher.valid_input_file_types,
                                                                          existing_file=existing_file)
            if input_file is not None:
                self.input_file_edit.setText(str(input_file.absolute()))
        else:
            self.patcher.delete_internal_copy(self._options.internal_copies_path)
            self.input_file_edit.setText("")
            self.check_extracted_game()

    # Output File
    def _validate_output_file(self):
        output_file = self.output_file
        has_error = output_file.is_dir() or not output_file.parent.is_dir()

        common_qt_lib.set_error_border_stylesheet(self.output_file_edit, has_error)
        self._update_accept_button()

    def _on_output_file_button(self):
        suggested_name = "{}.{}".format(self.default_output_name, self._selected_output_format)
        if self.output_file_edit.text() and self.output_file.parent.is_dir():
            suggested_name = str(self.output_file.parent.joinpath(suggested_name))

        output_file = common_qt_lib.prompt_user_for_output_file(self, suggested_name,
                                                                self.patcher.valid_output_file_types)
        if output_file is None:
            return

        output_file = output_file.absolute()

        try:
            output_file.parent.mkdir(parents=True, exist_ok=True)
        except (FileNotFoundError, OSError) as error:
            QMessageBox.warning(self,
                                "Invalid output file",
                                "Unable to use '{}' as output file: {}".format(output_file, error),
                                )
            return

        self.output_file_edit.setText(str(output_file))

    def _on_output_format_changed(self):
        button = self.sender()
        if button.isChecked():
            self._selected_output_format = button.text()[1:]
            current_filename = Path(self.output_file_edit.text())
            if str(current_filename) != '.':
                self.output_file_edit.setText(str(current_filename.with_suffix('.' + self._selected_output_format)))
                self._validate_output_file()



from typing import List

from PySide2.QtWidgets import QDialog

from randovania.game_description.world.world import World
from randovania.gui import game_specific_gui
from randovania.gui.generated.logic_settings_window_ui import Ui_LogicSettingsWindow
from randovania.gui.lib import common_qt_lib
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset

def dark_world_flags(world: World):
    yield False
    if world.dark_name is not None:
        yield True


class LogicSettingsWindow(QDialog, Ui_LogicSettingsWindow):
    _tabs: List[PresetTab]
    _editor: PresetEditor

    def __init__(self, window_manager: WindowManager, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self._editor = editor
        self._tabs = game_specific_gui.preset_editor_tabs_for(editor, window_manager)

        for extra_tab in self._tabs:
            if extra_tab.uses_patches_tab:
                parent_tab = self.patches_tab_widget
            else:
                parent_tab = self.logic_tab_widget
            parent_tab.addTab(extra_tab, extra_tab.tab_title)

        self.name_edit.textEdited.connect(self._edit_name)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

    # Options
    def on_preset_changed(self, preset: Preset):
        common_qt_lib.set_edit_if_different(self.name_edit, preset.name)
        for extra_tab in self._tabs:
            extra_tab.on_preset_changed(preset)

    def _edit_name(self, value: str):
        with self._editor as editor:
            editor.name = value



import asyncio

import pytest
from PySide2 import QtWidgets
from PySide2.QtWidgets import QDialog
from mock import AsyncMock, MagicMock, patch

from randovania.gui.main_online_interaction import OnlineInteractions
from randovania.interface_common.options import Options
from randovania.network_client.network_client import ConnectionState


@pytest.fixture(name="default_online_interactions")
def _default_online_interactions(skip_qtbot, preset_manager) -> OnlineInteractions:
    main_window = MagicMock()
    parent = QtWidgets.QWidget()
    skip_qtbot.add_widget(parent)

    return OnlineInteractions(parent, preset_manager, MagicMock(), main_window, Options(MagicMock()))


@pytest.mark.asyncio
@patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
async def test_browse_for_game_session(mock_execute_dialog: AsyncMock,
                                       skip_qtbot, default_online_interactions, mocker):
    # Setup
    mock_game_session_browser: MagicMock = mocker.patch(
        "randovania.gui.main_online_interaction.GameSessionBrowserDialog", autospec=True)
    mock_create_and_update: AsyncMock = mocker.patch(
        "randovania.gui.main_online_interaction.GameSessionWindow.create_and_update", new_callable=AsyncMock)
    mock_get_game_connection = mocker.patch("randovania.gui.lib.common_qt_lib.get_game_connection", autospec=True)
    default_online_interactions._ensure_logged_in = AsyncMock(return_value=True)
    mock_execute_dialog.return_value = mock_game_session_browser.return_value.Accepted
    mock_game_session_browser.return_value.refresh = AsyncMock()

    # Run
    await default_online_interactions._browse_for_game_session()

    # Assert
    mock_game_session_browser.assert_called_once_with(default_online_interactions.network_client)
    mock_game_session_browser.return_value.refresh.assert_awaited_once_with()
    mock_execute_dialog.assert_awaited_once_with(mock_game_session_browser.return_value)
    mock_create_and_update.assert_awaited_once_with(
        default_online_interactions.network_client,
        mock_get_game_connection.return_value,
        default_online_interactions.preset_manager,
        default_online_interactions.window_manager,
        default_online_interactions.options,
    )
    mock_create_and_update.return_value.show.assert_called_once_with()


@pytest.mark.asyncio
@patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
async def test_host_game_session(mock_execute_dialog: AsyncMock,
                                 skip_qtbot, default_online_interactions, mocker):
    # Setup
    mock_create_and_update: AsyncMock = mocker.patch(
        "randovania.gui.main_online_interaction.GameSessionWindow.create_and_update", new_callable=AsyncMock)
    mock_get_game_connection = mocker.patch("randovania.gui.lib.common_qt_lib.get_game_connection", autospec=True)
    default_online_interactions._ensure_logged_in = AsyncMock(return_value=True)
    mock_execute_dialog.return_value = QDialog.Accepted
    default_online_interactions.network_client.create_new_session = AsyncMock()

    # Run
    await default_online_interactions._host_game_session()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    default_online_interactions.network_client.create_new_session.assert_awaited_once_with("")
    mock_create_and_update.assert_awaited_once_with(
        default_online_interactions.network_client,
        mock_get_game_connection.return_value,
        default_online_interactions.preset_manager,
        default_online_interactions.window_manager,
        default_online_interactions.options,
    )
    mock_create_and_update.return_value.show.assert_called_once_with()


@pytest.mark.asyncio
async def test_ensure_logged_in(default_online_interactions, mocker):
    # Setup
    mock_message_box = mocker.patch("PySide2.QtWidgets.QMessageBox")

    async def true(): return True

    connect_task = asyncio.create_task(true())

    network_client = default_online_interactions.network_client
    network_client.connect_to_server.return_value = connect_task
    network_client.connection_state = ConnectionState.Disconnected
    network_client.current_user = MagicMock()

    # Run
    result = await default_online_interactions._ensure_logged_in()

    # Assert
    mock_message_box.assert_called_once()
    assert result



from typing import Optional

from randovania.interface_common import echoes
from randovania.interface_common.options import Options
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.lib.status_update_lib import ProgressUpdateCallable, ConstantPercentageCallback

export_busy = False


def generate_layout(options: Options,
                    permalink: Permalink,
                    progress_update: ProgressUpdateCallable,
                    retries: Optional[int] = None,
                    ) -> LayoutDescription:
    """
    Creates a LayoutDescription for the configured permalink
    :param options:
    :param permalink:
    :param progress_update:
    :param retries:
    :return:
    """
    return echoes.generate_description(
        permalink=permalink,
        status_update=ConstantPercentageCallback(progress_update, -1),
        validate_after_generation=options.advanced_validate_seed_after,
        timeout_during_generation=options.advanced_timeout_during_generation,
        attempts=retries,
    )



import dataclasses

from randovania.games.game import RandovaniaGame
from randovania.layout.base.base_configuration import BaseConfiguration
from randovania.layout.lib.teleporters import TeleporterConfiguration


@dataclasses.dataclass(frozen=True)
class CorruptionConfiguration(BaseConfiguration):
    elevators: TeleporterConfiguration
    energy_per_tank: int = dataclasses.field(metadata={"min": 1, "max": 1000, "precision": 1})
    start_with_corrupted_hypermode: bool = False

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        return RandovaniaGame.METROID_PRIME_CORRUPTION



import dataclasses
from typing import List

from PySide2 import QtWidgets

from randovania.game_description import default_database
from randovania.game_description.item.item_database import ItemDatabase
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.gui.lib import signal_handling
from randovania.gui.lib.scroll_protected import ScrollProtectedSpinBox
from randovania.gui.preset_settings.item_pool_tab import PresetItemPool
from randovania.gui.preset_settings.pickup_style_widget import PickupStyleWidget
from randovania.gui.preset_settings.progressive_item_widget import ProgressiveItemWidget
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.major_item_state import DEFAULT_MAXIMUM_SHUFFLED
from randovania.layout.preset import Preset


class MetroidPresetItemPool(PresetItemPool):
    _progressive_widgets: List[ProgressiveItemWidget]

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        item_database = default_database.item_database_for_game(self.game)
        game_description = default_database.game_description_for(self.game)

        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)

        self._energy_tank_item = item_database.major_items["Energy Tank"]
        self._create_energy_tank_box(game_description.resource_database.energy_tank)
        self._create_progressive_widgets(item_database)
        self._create_pickup_style_box(size_policy)
        signal_handling.on_checked(self.multi_pickup_placement_check, self._persist_multi_pickup_placement)

    def on_preset_changed(self, preset: Preset):
        super().on_preset_changed(preset)
        layout = preset.configuration
        major_configuration = layout.major_items_configuration

        for progressive_widget in self._progressive_widgets:
            progressive_widget.on_preset_changed(
                preset,
                self._boxes_for_category[progressive_widget.progressive_item.item_category.name][2],
            )

        self.pickup_style_widget.update(layout)

        # Energy Tank
        energy_tank_state = major_configuration.items_state[self._energy_tank_item]
        self.energy_tank_starting_spinbox.setValue(energy_tank_state.num_included_in_starting_items)
        self.energy_tank_shuffled_spinbox.setValue(energy_tank_state.num_shuffled_pickups)

    def _persist_multi_pickup_placement(self, value: bool):
        with self._editor as editor:
            editor.set_configuration_field("multi_pickup_placement", value)

    def _create_pickup_style_box(self, size_policy):
        self.pickup_style_widget = PickupStyleWidget(None, self._editor)
        self.item_pool_layout.insertWidget(1, self.pickup_style_widget)
        # widget.Changed.connect(partial(self._on_major_item_updated, widget))

    def _create_energy_tank_box(self, energy_tank_resource: ItemResourceInfo):
        category_box, category_layout, _ = self._boxes_for_category["energy_tank"]

        starting_label = QtWidgets.QLabel(category_box)
        starting_label.setText("Starting Quantity")
        category_layout.addWidget(starting_label, 0, 0)

        self.energy_tank_starting_spinbox = ScrollProtectedSpinBox(category_box)
        self.energy_tank_starting_spinbox.setMaximum(energy_tank_resource.max_capacity)
        self.energy_tank_starting_spinbox.valueChanged.connect(self._on_update_starting_energy_tank)
        category_layout.addWidget(self.energy_tank_starting_spinbox, 0, 1)

        shuffled_label = QtWidgets.QLabel(category_box)
        shuffled_label.setText("Shuffled Quantity")
        category_layout.addWidget(shuffled_label, 1, 0)

        self.energy_tank_shuffled_spinbox = ScrollProtectedSpinBox(category_box)
        self.energy_tank_shuffled_spinbox.setMaximum(DEFAULT_MAXIMUM_SHUFFLED[-1])
        self.energy_tank_shuffled_spinbox.valueChanged.connect(self._on_update_shuffled_energy_tank)
        category_layout.addWidget(self.energy_tank_shuffled_spinbox, 1, 1)

    def _on_update_starting_energy_tank(self, value: int):
        with self._editor as options:
            major_configuration = options.major_items_configuration
            options.major_items_configuration = major_configuration.replace_state_for_item(
                self._energy_tank_item,
                dataclasses.replace(major_configuration.items_state[self._energy_tank_item],
                                    num_included_in_starting_items=value)
            )

    def _on_update_shuffled_energy_tank(self, value: int):
        with self._editor as options:
            major_configuration = options.major_items_configuration
            options.major_items_configuration = major_configuration.replace_state_for_item(
                self._energy_tank_item,
                dataclasses.replace(major_configuration.items_state[self._energy_tank_item],
                                    num_shuffled_pickups=value)
            )

    def _create_progressive_widgets(self, item_database: ItemDatabase):
        self._progressive_widgets = []

        all_progressive = []

        if self.game == RandovaniaGame.METROID_PRIME_ECHOES:
            all_progressive.append((
                "Progressive Suit",
                ("Dark Suit", "Light Suit"),
            ))
            all_progressive.append((
                "Progressive Grapple",
                ("Grapple Beam", "Screw Attack"),
            ))
        elif self.game == RandovaniaGame.METROID_PRIME_CORRUPTION:
            all_progressive.append((
                "Progressive Missile",
                ("Ice Missile", "Seeker Missile"),
            ))
            all_progressive.append((
                "Progressive Beam",
                ("Plasma Beam", "Nova Beam"),
            ))

        for (progressive_item_name, non_progressive_items) in all_progressive:
            progressive_item = item_database.major_items[progressive_item_name]
            parent, layout, _ = self._boxes_for_category[progressive_item.item_category.name]

            widget = ProgressiveItemWidget(
                parent, self._editor,
                progressive_item=progressive_item,
                non_progressive_items=[item_database.major_items[it] for it in non_progressive_items],
            )
            widget.setText("Use progressive {}".format(" → ".join(non_progressive_items)))
            self._progressive_widgets.append(widget)
            layout.addWidget(widget)



"""
Some constants related to specific Corruption items that the logic needs to deal with.
"""

ENERGY_CELL_MODEL = "Energy Cell"
ENERGY_CELL_ITEMS = [23 + i for i in range(9)]
ENERGY_CELL_TOTAL_ITEM = 22



from argparse import ArgumentParser


def add_debug_argument(parser: ArgumentParser):
    parser.add_argument(
        "--debug",
        choices=range(4),
        type=int,
        default=0,
        help="The level of debug logging to print.")


def add_validate_argument(parser: ArgumentParser):
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--validate", action="store_true", dest="validate", default=True,
                       help="After generating a layout, validate if it's possible. Default behaviour.")
    group.add_argument("--no-validate", action="store_false", dest="validate", default=True,
                       help="After generating a layout, don't validate if it's possible.")



import dataclasses
from random import Random
from typing import List, Dict

from randovania.game_description.assignment import PickupTarget
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import (PickupEntry, ConditionalResources, PickupModel,
                                                                ResourceConversion)
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import ResourceGainTuple
from randovania.game_description.world.node import PickupNode
from randovania.game_description.world.world_list import WorldList
from randovania.games.prime.patcher_file_lib import item_names
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.base.pickup_model import PickupModelStyle, PickupModelDataSource


def _conditional_resources_for_pickup(pickup: PickupEntry) -> List[ConditionalResources]:
    if len(pickup.progression) > 1:
        assert pickup.resource_lock is None, pickup.name
        return list(pickup.conditional_resources)

    else:
        resources = list(pickup.extra_resources)
        name = pickup.name
        if pickup.progression:
            name = pickup.progression[0][0].long_name
            resources.append(pickup.progression[0])

        lock = pickup.resource_lock
        if pickup.respects_lock and not pickup.unlocks_resource and lock is not None:
            locked_resources = lock.convert_gain(resources)
            return [
                ConditionalResources(
                    name=f"Locked {name}",
                    item=None,
                    resources=tuple(locked_resources),
                ),
                ConditionalResources(
                    name=name,
                    item=lock.locked_by,
                    resources=tuple(resources),
                ),
            ]
        else:
            return [
                ConditionalResources(
                    name=name,
                    item=None,
                    resources=tuple(resources),
                ),
            ]


def _pickup_scan(pickup: PickupEntry) -> str:
    if not pickup.item_category.is_expansion:
        if len(pickup.progression) > 1:
            return "{}. Provides the following in order: {}".format(
                pickup.name, ", ".join(conditional.name for conditional in pickup.conditional_resources))
        else:
            return pickup.name

    ammo_desc = [
        item_names.add_quantity_to_resource(item_names.resource_user_friendly_name(resource), quantity, True)
        for resource, quantity in pickup.extra_resources
    ]
    if ammo_desc:
        return "{}. Provides {}{}{}".format(
            pickup.name,
            ", ".join(ammo_desc[:-1]),
            " and " if len(ammo_desc) > 1 else "",
            ammo_desc[-1],
        )
    else:
        return pickup.name


def _get_single_hud_text(pickup_name: str,
                         memo_data: Dict[str, str],
                         resources: ResourceGainTuple,
                         ) -> str:
    return memo_data[pickup_name].format(**{
        item_names.resource_user_friendly_name(resource): quantity
        for resource, quantity in resources
    })


def _get_all_hud_text(conditionals: List[ConditionalResources],
                      memo_data: Dict[str, str],
                      ) -> List[str]:
    return [
        _get_single_hud_text(conditional.name, memo_data, conditional.resources)
        for conditional in conditionals
    ]


def _calculate_hud_text(pickup: PickupEntry,
                        visual_pickup: PickupEntry,
                        model_style: PickupModelStyle,
                        memo_data: Dict[str, str],
                        ) -> List[str]:
    """
    Calculates what the hud_text for a pickup should be
    :param pickup:
    :param visual_pickup:
    :param model_style:
    :param memo_data:
    :return:
    """

    if model_style == PickupModelStyle.HIDE_ALL:
        hud_text = _get_all_hud_text(_conditional_resources_for_pickup(visual_pickup), memo_data)
        num_conditional = len(_conditional_resources_for_pickup(pickup))
        if len(hud_text) == num_conditional:
            return hud_text
        else:
            return [hud_text[0]] * num_conditional

    else:
        return _get_all_hud_text(_conditional_resources_for_pickup(pickup), memo_data)


@dataclasses.dataclass(frozen=True)
class ExportedPickupDetails:
    index: PickupIndex
    scan_text: str
    hud_text: List[str]
    conditional_resources: List[ConditionalResources]
    conversion: List[ResourceConversion]
    model: PickupModel
    other_player: bool


class PickupExporter:
    def create_details(self,
                       original_index: PickupIndex,
                       pickup_target: PickupTarget,
                       visual_pickup: PickupEntry,
                       model_style: PickupModelStyle,
                       scan_text: str,
                       model: PickupModel) -> ExportedPickupDetails:
        raise NotImplementedError()

    def export(self,
               original_index: PickupIndex,
               pickup_target: PickupTarget,
               visual_pickup: PickupEntry,
               model_style: PickupModelStyle,
               ) -> ExportedPickupDetails:
        model_pickup = pickup_target.pickup if model_style == PickupModelStyle.ALL_VISIBLE else visual_pickup

        if model_style in {PickupModelStyle.ALL_VISIBLE, PickupModelStyle.HIDE_MODEL}:
            scan_text = _pickup_scan(pickup_target.pickup)
        else:
            scan_text = visual_pickup.name

        return self.create_details(original_index, pickup_target, visual_pickup,
                                   model_style, scan_text, model_pickup.model)


class PickupExporterSolo(PickupExporter):
    def __init__(self, memo_data: Dict[str, str]):
        self.memo_data = memo_data

    def create_details(self,
                       original_index: PickupIndex,
                       pickup_target: PickupTarget,
                       visual_pickup: PickupEntry,
                       model_style: PickupModelStyle,
                       scan_text: str,
                       model: PickupModel) -> ExportedPickupDetails:
        pickup = pickup_target.pickup
        return ExportedPickupDetails(
            index=original_index,
            scan_text=scan_text,
            hud_text=_calculate_hud_text(pickup, visual_pickup, model_style, self.memo_data),
            conditional_resources=_conditional_resources_for_pickup(pickup),
            conversion=list(pickup.convert_resources),
            model=model,
            other_player=False,
        )


class PickupExporterMulti(PickupExporter):
    def __init__(self, solo_creator: PickupExporter, multiworld_item: ItemResourceInfo,
                 players_config: PlayersConfiguration):
        self.solo_creator = solo_creator
        self.multiworld_item = multiworld_item
        self.players_config = players_config

    def create_details(self,
                       original_index: PickupIndex,
                       pickup_target: PickupTarget,
                       visual_pickup: PickupEntry,
                       model_style: PickupModelStyle,
                       scan_text: str,
                       model: PickupModel) -> ExportedPickupDetails:
        if pickup_target.player == self.players_config.player_index:
            details = self.solo_creator.create_details(original_index, pickup_target, visual_pickup,
                                                       model_style, scan_text, model)
            return dataclasses.replace(details, scan_text=f"Your {details.scan_text}")
        else:
            other_name = self.players_config.player_names[pickup_target.player]
            return ExportedPickupDetails(
                index=original_index,
                scan_text=f"{other_name}'s {scan_text}",
                hud_text=[f"Sent {pickup_target.pickup.name} to {other_name}!"],
                conditional_resources=[ConditionalResources(
                    name=None,
                    item=None,
                    resources=(
                        (self.multiworld_item, original_index.index + 1),
                    ),
                )],
                conversion=[],
                model=model,
                other_player=True,
            )


def _get_visual_model(original_index: int,
                      pickup_list: List[PickupTarget],
                      data_source: PickupModelDataSource,
                      visual_etm: PickupEntry,
                      ) -> PickupEntry:
    if data_source == PickupModelDataSource.ETM:
        return visual_etm
    elif data_source == PickupModelDataSource.RANDOM:
        return pickup_list[original_index % len(pickup_list)].pickup
    elif data_source == PickupModelDataSource.LOCATION:
        raise NotImplementedError()
    else:
        raise ValueError(f"Unknown data_source: {data_source}")


def export_all_indices(patches: GamePatches,
                       useless_target: PickupTarget,
                       world_list: WorldList,
                       rng: Random,
                       model_style: PickupModelStyle,
                       data_source: PickupModelDataSource,
                       exporter: PickupExporter,
                       visual_etm: PickupEntry,
                       ) -> List[ExportedPickupDetails]:
    """
    Creates the patcher data for all pickups in the game
    :param patches:
    :param useless_target:
    :param world_list:
    :param rng:
    :param model_style:
    :param data_source:
    :param exporter:
    :param visual_etm:
    :return:
    """
    pickup_assignment = patches.pickup_assignment

    pickup_list = list(pickup_assignment.values())
    rng.shuffle(pickup_list)

    indices = sorted(
        node.pickup_index
        for node in world_list.all_nodes
        if isinstance(node, PickupNode)
    )

    pickups = [
        exporter.export(index,
                        pickup_assignment.get(index, useless_target),
                        _get_visual_model(i, pickup_list, data_source, visual_etm),
                        model_style,
                        )
        for i, index in enumerate(indices)
    ]

    return pickups


class GenericAcquiredMemo(dict):
    def __missing__(self, key):
        return "{} acquired!".format(key)


def create_pickup_exporter(game: GameDescription, memo_data: dict, players_config: PlayersConfiguration):
    exporter = PickupExporterSolo(memo_data)
    if players_config.is_multiworld:
        exporter = PickupExporterMulti(exporter, game.resource_database.multiworld_magic_item,
                                       players_config)
    return exporter



import copy
from dataclasses import dataclass
from typing import Dict, Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.game_description import default_database
from randovania.game_description.item.ammo import Ammo
from randovania.games.game import RandovaniaGame
from randovania.layout.base.ammo_state import AmmoState


@dataclass(frozen=True)
class AmmoConfiguration(bitpacking.BitPackValue):
    items_state: Dict[Ammo, AmmoState]

    def __post_init__(self):
        for ammo, state in self.items_state.items():
            state.check_consistency(ammo)

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        default: AmmoConfiguration = metadata["reference"]

        assert list(self.items_state.keys()) == list(default.items_state.keys())

        for ammo, this in self.items_state.items():
            reference = default.items_state[ammo]
            is_different = this != reference
            yield from bitpacking.encode_bool(is_different)
            if is_different:
                yield from this.bit_pack_encode({
                    "ammo": ammo,
                })

    @classmethod
    def bit_pack_unpack(cls, decoder: bitpacking.BitPackDecoder, metadata):
        default: AmmoConfiguration = metadata["reference"]

        items_state = {}
        for ammo, default_state in default.items_state.items():
            is_different = bitpacking.decode_bool(decoder)
            if is_different:
                items_state[ammo] = AmmoState.bit_pack_unpack(decoder, {"ammo": ammo})
            else:
                items_state[ammo] = default_state

        return cls(items_state)

    @property
    def as_json(self) -> dict:
        return {
            "items_state": {
                ammo.name: state.as_json
                for ammo, state in self.items_state.items()
            },
        }

    @classmethod
    def from_json(cls, value: dict, game: RandovaniaGame) -> "AmmoConfiguration":
        item_database = default_database.item_database_for_game(game)
        return cls(
            items_state={
                item_database.ammo[name]: AmmoState.from_json(state)
                for name, state in value["items_state"].items()
            },
        )

    def replace_state_for_ammo(self, ammo: Ammo, state: AmmoState) -> "AmmoConfiguration":
        return self.replace_states({ammo: state})

    def replace_states(self, new_states: Dict[Ammo, AmmoState]) -> "AmmoConfiguration":
        """
        Creates a copy of this AmmoConfiguration where the state of all given items are replaced by the given
        states.
        :param new_states:
        :return:
        """
        items_state = copy.copy(self.items_state)

        for item, state in new_states.items():
            items_state[item] = state

        return AmmoConfiguration(items_state)



import dulwich.repo
import pytest

from randovania.interface_common import preset_manager
from randovania.layout.preset_migration import VersionedPreset


@pytest.mark.asyncio
async def test_add_then_delete_preset(tmp_path, default_preset):
    p = VersionedPreset.with_preset(default_preset.fork())

    dulwich.repo.Repo.init(tmp_path)
    manager = preset_manager.PresetManager(tmp_path.joinpath("presets"))
    await manager.load_user_presets()

    assert manager.preset_for_uuid(p.uuid) is None
    manager.add_new_preset(p)
    assert manager.preset_for_uuid(p.uuid) == p
    manager.delete_preset(p)
    assert manager.preset_for_uuid(p.uuid) is None

